"""Workflows to derive river for a wflow model."""

import logging
from os.path import join

import geopandas as gpd
import numpy as np
import pandas as pd
import pyflwdir
import xarray as xr
from hydromt import stats
from hydromt.gis import flw
from hydromt.gis.raster_utils import cellres
from hydromt.gis.vector_utils import nearest
from hydromt.model.processes.rivers import river_depth
from scipy.optimize import curve_fit

from hydromt_wflow.utils import DATADIR  # global var

logger = logging.getLogger(f"hydromt.{__name__}")

__all__ = ["river", "river_bathymetry", "river_width", "river_floodplain_volume"]

RESAMPLING = {"climate": "nearest", "precip": "average"}
NODATA = {"discharge": -9999}


def power_law(x, a, b):
    return a * np.power(x, b)


def river(
    ds: xr.Dataset,
    ds_model: xr.Dataset | None = None,
    river_upa: float = 30.0,
    slope_len: float = 2e3,
    min_rivlen_ratio: float = 0.0,
    channel_dir: str = "up",
):
    """Return river maps.

    The output maps are:

    - rivmsk : river mask based on upstream area threshold on upstream area
    - rivlen : river length [m]
    - rivslp : smoothed river slope [m/m]
    - rivzs : elevation of the river bankfull height based on pixel outlet
    - rivwth : river width at pixel outlet (if in ds)
    - qbankfull : bankfull discharge at pixel outlet (if in ds)

    Parameters
    ----------
    ds: xr.Dataset
        hydrography dataset containing "flwdir", "uparea", "elevtn" variables;
        and optional "rivwth" and "qbankfull" variable
    ds_model: xr.Dataset, optional
        model dataset with output grid, must contain "uparea", for subgrid rivlen/slp
        must contain "x_out", "y_out". If None, ds is assumed to be the model grid
    river_upa: float
        minimum threshold to define river cell & pixels, by default 30 [km2]
    slope_len: float
        minimum length over which to calculate the river slope, by default 1000 [m]
    min_rivlen_ratio: float
        minimum global river length to avg. Cell resolution ratio used as threshold
        in window based smoothing of river length, by default 0.0.
        The smoothing is skipped if min_riverlen_ratio = 0.
    channel_dir: {"up", "down"}
        flow direction in which to calculate (subgrid) river length and width

    Returns
    -------
    ds_out: xr.Dataset
        Dataset with output river attributes
    """
    # check data variables.
    dvars = ["flwdir", "uparea", "elevtn"]
    dvars_model = ["flwdir", "uparea"]
    if not np.all([v in ds for v in dvars]):
        raise ValueError(f"One or more variables missing from ds: {dvars}.")
    if ds_model is not None and not np.all([v in ds_model for v in dvars_model]):
        raise ValueError(f"One or more variables missing from ds_model: {dvars_model}")
    # sort subgrid
    subgrid = True
    if ds_model is None or not np.all([v in ds_model for v in ["x_out", "y_out"]]):
        subgrid = False
        if ds_model is not None:
            ds = ds.rename(
                {
                    ds.raster.y_dim: ds_model.raster.y_dim,
                    ds.raster.x_dim: ds_model.raster.x_dim,
                }
            )
            try:
                ds = ds.sel({dim: ds_model[dim].values for dim in ds_model.raster.dims})
            except KeyError:
                raise ValueError(
                    "It seems model grid was not upscaled during setup_basemaps however"
                    " model grid does not align with the hydrography data provided for"
                    " river processing. Make sure you are using the same hydrography"
                    " dataset for setup_basemaps and setup_rivers."
                )
        ds_model = ds
        logger.info("River length and slope are calculated at model resolution.")

    ## river mask and flow directions at model grid
    logger.info(f"Set river mask (min uparea: {river_upa} km2) and prepare flow dirs.")
    riv_mask = ds_model["uparea"] > river_upa  # initial mask
    mod_mask = ds_model["uparea"] != ds_model["uparea"].raster.nodata

    ## (high res) flwdir and outlet indices
    riv_mask_org = ds["uparea"].values >= river_upa  # highres riv mask
    flwdir = flw.flwdir_from_da(ds["flwdir"], mask=True)
    if subgrid == False:
        # get cell index of river cells
        idxs_out = np.arange(ds_model.raster.size).reshape(ds_model.raster.shape)
        idxs_out = np.where(mod_mask, idxs_out, flwdir._mv)
    else:
        # get subgrid outlet pixel index
        idxs_out = ds.raster.xy_to_idx(
            xs=ds_model["x_out"].values,
            ys=ds_model["y_out"].values,
            mask=mod_mask.values,
            nodata=flwdir._mv,
        )

    ## river length
    # get river length based on on-between distance between two outlet pixels
    logger.info("Derive river length.")
    rivlen = flwdir.subgrid_rivlen(
        idxs_out=idxs_out,
        mask=riv_mask_org,
        direction=channel_dir,
        unit="m",
    )
    xres, yres = ds_model.raster.res
    if ds_model.raster.crs.is_geographic:  # convert degree to meters
        lat_avg = ds_model.raster.ycoords.values.mean()
        xres, yres = cellres(lat_avg, xres, yres)
    rivlen = np.where(riv_mask.values, rivlen, -9999)
    # set mean length at most downstream (if channel_dir=down) or upstream
    # (if channel_dir=up) river lengths
    if np.any(rivlen == 0):
        rivlen[rivlen == 0] = np.mean(rivlen[rivlen > 0])
    # smooth river length based on minimum river length
    if min_rivlen_ratio > 0 and hasattr(flwdir, "smooth_rivlen"):
        res = np.mean(np.abs([xres, yres]))
        min_len = res * min_rivlen_ratio
        flwdir_model = flw.flwdir_from_da(ds_model["flwdir"], mask=riv_mask)
        rivlen2 = flwdir_model.smooth_rivlen(rivlen, min_len, nodata=-9999)
        min_len2 = rivlen2[riv_mask].min()
        pmod = (rivlen != rivlen2).sum() / riv_mask.sum() * 100
        logger.info(
            f"River length smoothed (min length: {min_len2:.0f} m; \
cells modified: {pmod:.1f})%."
        )
        rivlen = rivlen2
    elif min_rivlen_ratio > 0:
        logger.warning(
            "River length smoothing skipped as it requires newer version of pyflwdir."
        )

    ## river slope as derivative of elevation around outlet pixels
    logger.info("Derive river slope.")
    rivslp = flwdir.subgrid_rivslp(
        idxs_out=idxs_out,
        elevtn=ds["elevtn"].values,
        mask=riv_mask_org,
        length=slope_len,
    )
    rivslp = np.where(riv_mask.values, rivslp, -9999)

    # create xarray dataset for all river variables
    ds_out = xr.Dataset(coords=ds_model.raster.coords)
    dims = ds_model.raster.dims
    # save as uint8 as bool is not supported in nc and tif files
    ds_out["rivmsk"] = riv_mask.astype(np.uint8)
    ds_out["rivmsk"].raster.set_nodata(0)
    attrs = dict(_FillValue=-9999, unit="m")
    ds_out["rivlen"] = xr.Variable(dims, rivlen, attrs=attrs)
    attrs = dict(_FillValue=-9999, unit="m.m-1")
    ds_out["rivslp"] = xr.Variable(dims, rivslp, attrs=attrs)

    for name in ["rivwth", "qbankfull"]:
        if name in ds:
            logger.debug(f"Derive {name} from hydrography dataset.")
            data = np.full_like(riv_mask, -9999, dtype=np.float32)
            data0 = ds[name].values.flat[idxs_out[riv_mask.values]]
            data[riv_mask.values] = np.where(data0 > 0, data0, -9999)
            ds_out[name] = xr.Variable(dims, data, attrs=dict(_FillValue=-9999))

    return ds_out, flwdir


def river_bathymetry(
    ds_model: xr.Dataset,
    gdf_riv: gpd.GeoDataFrame,
    method: str | None = "powlaw",
    smooth_len: float = 5e3,
    min_rivdph: float = 1.0,
    min_rivwth: float = 30.0,
) -> xr.Dataset:
    """Get river width and optionally river depth/bankfull discharge.

    If `method` is None, only river width is derived. Otherwise river depth
    is estimated using :py:meth:`hydromt.workflows.river_depth`.

    Parameters
    ----------
    ds_model : xr.Dataset
        Model dataset with 'flwdir', 'rivmsk', 'rivlen', 'x_out' and 'y_out' variables.
    gdf_riv : gpd.GeoDataFrame
        River geometry with 'rivwth' and optionally 'rivdph' or 'qbankfull' columns.
    method : {'gvf', 'manning', 'powlaw', None}, optional
        Method to compute river depth. If None, depth is skipped.
    smooth_len : float, optional
        Length [m] over which to smooth the output river width and depth, by default 5e3
    min_rivdph : float, optional
        Minimum river depth [m], by default 1.0. Ignored if method=None.
    min_rivwth : float, optional
        Minimum river width [m], by default 30.0.

    Returns
    -------
    xr.Dataset
        Dataset with 'rivwth' and (if method not None) 'rivdph'.
    """
    dims = ds_model.raster.dims

    # check required model variables
    dvars_model = ["flwdir", "rivmsk", "rivlen"]
    if method == "manning":
        dvars_model += ["rivslp"]
    elif method == "gvf":
        dvars_model += ["rivslp", "subelv"]
    if not np.all([v in ds_model for v in dvars_model]):
        raise ValueError(f"One or more variables missing from ds_model: {dvars_model}")

    # setup flow direction for river
    riv_mask = ds_model["rivmsk"].values == 1
    flwdir_river = flw.flwdir_from_da(ds_model["flwdir"], mask=riv_mask)
    rivlen_avg = ds_model["rivlen"].values[riv_mask].mean()

    ## river width (+ optional depth)
    vars0 = ["rivwth"]
    if method is not None:
        vars0 += ["rivdph", "qbankfull"]

    # find nearest values from river shape if provided
    # if None assume the data is in ds_model
    if gdf_riv is not None:
        vars = [c for c in vars0 if c in gdf_riv.columns]
        if len(vars) == 0:
            raise ValueError(f"columns {vars0} not found in gdf_riv")
        logger.info(f"Derive {vars} from shapefile.")

        # outlet points
        if "x_out" in ds_model and "y_out" in ds_model:
            # get subgrid outlet pixel index and coordinates
            xs_out = ds_model["x_out"].values[riv_mask]
            ys_out = ds_model["y_out"].values[riv_mask]
        else:
            # get river cell coordinates
            row, col = np.nonzero(riv_mask)
            xs_out = ds_model.raster.xcoords.values[col]
            ys_out = ds_model.raster.ycoords.values[row]

        gdf_out = gpd.GeoDataFrame(
            geometry=gpd.points_from_xy(xs_out, ys_out), crs=ds_model.raster.crs
        )
        idx_nn, dst_nn = nearest(gdf_out, gdf_riv)

        # get valid river data within max half pixel distance
        xres, yres = ds_model.raster.res
        if ds_model.raster.crs.is_geographic:  # convert degree to meters
            lat_avg = ds_model.raster.ycoords.values.mean()
            xres, yres = cellres(lat_avg, xres, yres)
        max_dist = np.mean(np.abs([xres, yres])) / 2.0

        nriv, nsnap = xs_out.size, int(np.sum(dst_nn < max_dist))
        logger.info(
            f"Valid for {nsnap}/{nriv} river cells (max dist: {max_dist:.0f} m)."
        )

        for name in vars:
            data = np.full_like(riv_mask, -9999, dtype=np.float32)
            data[riv_mask] = np.where(
                dst_nn < max_dist, gdf_riv.loc[idx_nn, name].fillna(-9999).values, -9999
            )
            ds_model[name] = xr.Variable(dims, data, attrs=dict(_FillValue=-9999))
    else:
        vars = vars0

    assert "rivwth" in ds_model
    if method is not None:
        assert "qbankfull" in ds_model or "rivdph" in ds_model

    # fill gaps in data using downward filling along flow directions
    for name in vars:
        data = ds_model[name].values
        nodata = ds_model[name].raster.nodata
        if np.all(data[riv_mask] != nodata):
            continue
        data = flwdir_river.fillnodata(data, nodata, direction="down", how="max")
        ds_model[name].values = np.maximum(0, data)

    # smooth by averaging along flow directions and set minimum
    if smooth_len > 0:
        nsmooth = min(1, int(round(smooth_len / rivlen_avg / 2)))
        kwgs = dict(n=nsmooth, restrict_strord=True)
        ds_model["rivwth"].values = flwdir_river.moving_average(
            ds_model["rivwth"].values, nodata=ds_model["rivwth"].raster.nodata, **kwgs
        )
    ds_model["rivwth"] = np.maximum(min_rivwth, ds_model["rivwth"]).where(
        riv_mask, ds_model["rivwth"].raster.nodata
    )

    # optional depth
    if method is not None:
        if "rivdph" not in ds_model:
            # distance to outlet; required for manning and gvf rivdph methods
            if method != "powlaw" and "rivdst" not in ds_model:
                rivlen = ds_model["rivlen"].values
                nodata = ds_model["rivlen"].raster.nodata
                rivdst = flwdir_river.accuflux(rivlen, nodata=nodata, direction="down")
                ds_model["rivdst"] = xr.Variable(
                    dims, rivdst, attrs=dict(_FillValue=nodata)
                )
            # add river distance to outlet -> required for manning/gvf method
            rivdph = river_depth(
                data=ds_model,
                flwdir=flwdir_river,
                method=method,
                min_rivdph=min_rivdph,
                rivzs_name="subelv",
            )
            attrs = dict(_FillValue=-9999, unit="m")
            ds_model["rivdph"] = xr.Variable(dims, rivdph, attrs=attrs).fillna(-9999)

        # smooth by averaging along flow directions and set minimum
        if smooth_len > 0:
            ds_model["rivdph"].values = flwdir_river.moving_average(
                ds_model["rivdph"].values,
                nodata=-9999,
                **kwgs,
            )
        ds_model["rivdph"] = np.maximum(min_rivdph, ds_model["rivdph"]).where(
            riv_mask, -9999
        )

    # return only what was computed
    out_vars = ["rivwth"]
    if method is not None:
        out_vars.append("rivdph")
    return ds_model[out_vars]


def river_floodplain_volume(
    ds: xr.Dataset,
    ds_model: xr.Dataset,
    river_upa: float = 30,
    flood_depths: list[float] = [0.5, 1.0, 1.5, 2.0, 2.5],
    dtype: np.dtype = np.float64,
):
    """Calculate floodplain volume at given flood depths based on (subgrid) HAND map.

    Parameters
    ----------
    ds: xr.Dataset
        hydrography dataset containing "flwdir", "uparea", "elevtn" variables;
    ds_model: xr.Dataset, optional
        Model dataset with output grid, must contain "rivmsk", "rivwth", "rivlen"
        for subgrid must contain "x_out", "y_out".
    river_upa: float
        minimum threshold to define the river when calculating HAND, by default 30 [km2]
    flood_depths : list of float, optional
        flood depths at which a volume is derived, by default [0.5,1.0,1.5,2.0,2.5]
    dtype: numpy.dtype, optional
        output dtype

    Returns
    -------
    da_out : xr.DataArray with dims (flood_depth, y, x)
        Floodplain volume [m3]
    """
    if not ds.raster.crs == ds_model.raster.crs:
        raise ValueError("Hydrography dataset CRS does not match model CRS")

    # check data variables.
    dvars = ["flwdir", "elevtn", "uparea"]
    dvars_model = ["rivmsk", "rivwth", "rivlen"]
    if not np.all([v in ds for v in dvars]):
        raise ValueError(f"One or more variables missing from ds: {dvars}.")
    if not np.all([v in ds_model for v in dvars_model]):
        raise ValueError(f"One or more variables missing from ds_model: {dvars_model}")

    # initialize flood directions
    flwdir = flw.flwdir_from_da(ds["flwdir"], mask=None)

    # get river cell coordinates
    riv_mask = ds_model["rivmsk"].values > 0
    if "x_out" in ds_model and "y_out" in ds_model:  # use subgrid
        idxs_out = ds.raster.xy_to_idx(
            xs=ds_model["x_out"].values,
            ys=ds_model["y_out"].values,
            mask=riv_mask,
            nodata=flwdir._mv,
        )
    else:
        # get cell index of river cells
        idxs_out = np.arange(ds_model.raster.size).reshape(ds_model.raster.shape)
        idxs_out = np.where(riv_mask, idxs_out, flwdir._mv)

    # calculate hand
    hand = flwdir.hand(
        drain=ds["uparea"].values >= river_upa, elevtn=ds["elevtn"].values
    )

    # calculate volume at user defined flood depths
    # note that the ucat_map is not save currently but is required if
    # we want to create a flood map by postprocessing
    flood_depths = np.atleast_1d(flood_depths).ravel().astype(dtype)  # force 1D
    _, ucat_vol = flwdir.ucat_volume(idxs_out=idxs_out, hand=hand, depths=flood_depths)
    # force minimum volume based on river width * length
    rivlen = np.maximum(ds_model["rivlen"].values, 1)  # avoid zero division errors
    min_vol = ds_model["rivwth"].values * rivlen * flood_depths[0]
    ucat_vol[0, :, :] = np.maximum(ucat_vol[0, :, :], min_vol)
    fldwth = ucat_vol[0, :, :] / rivlen / flood_depths[0]
    for i in range(1, ucat_vol.shape[0]):
        dh = flood_depths[i] - flood_depths[i - 1]
        min_vol = ucat_vol[i - 1, ...] + (fldwth * rivlen * dh)
        ucat_vol[i, :, :] = np.maximum(ucat_vol[i, :, :], min_vol)
        fldwth = (ucat_vol[i, :, :] - ucat_vol[i - 1, :, :]) / rivlen / dh

    # return xarray DataArray
    da_out = (
        xr.DataArray(
            coords={"flood_depth": flood_depths, **ds_model.raster.coords},
            dims=("flood_depth", *ds_model.raster.dims),
            data=ucat_vol,
        )
        .reset_coords(drop=True)
        .where(ds_model["rivmsk"] > 0, -9999.0)
    )
    da_out.raster.set_nodata(-9999.0)
    da_out.raster.set_crs(ds_model.raster.crs)

    # TODO return a second dataset with hand and ucat_map variables for postprocessing
    return da_out


# TODO: methods below are redundant after version v0.1.4 and will be removed in
# future versions


def river_width(
    ds_like: xr.Dataset,
    flwdir: pyflwdir.FlwdirRaster,
    data: dict = {},
    fit: bool = False,
    fill: bool = True,
    fill_outliers: bool = True,
    min_wth: float = 1,
    mask_names: list = [],
    predictor: str = "discharge",
    rivwth_name: str = "rivwth",
    obs_postfix: str = "_obs",
    rivmsk_name: str = "rivmsk",
    a: float | None = None,
    b: float | None = None,
):
    """Calculate river width."""
    nopars = a is None or b is None  # no manual a, b parameters
    fit = fit or (nopars and predictor not in ["discharge"])  # fit power-law on the fly
    nowth = f"{rivwth_name}{obs_postfix}" not in ds_like  # no observed width
    fill = fill and nowth == False  # fill datagaps and masked areas (lakes/res) in obs

    if nowth and fit:
        raise ValueError(
            f'Observed rivwth "{rivwth_name}{obs_postfix}" required to fit riverwidths.'
        )
    elif nowth:
        logger.warning(f'Observed rivwth "{rivwth_name}{obs_postfix}" not found.')

    # get predictor
    logger.debug(f'Deriving predictor "{predictor}" values')
    if predictor == "discharge":
        values, pars = _discharge(ds_like, flwdir=flwdir, **data)
        if fit == False:
            a, b = pars
    # TODO: check units
    elif predictor == "precip":
        values = _precip(ds_like, flwdir=flwdir, **data)
    else:
        if predictor not in ds_like:
            raise ValueError(f"required {predictor} variable missing in grid.")
        values = ds_like[predictor].values

    # read river width observations
    if nowth == False:
        rivwth_org = ds_like[f"{rivwth_name}{obs_postfix}"].values
        nodata = ds_like[f"{rivwth_name}{obs_postfix}"].raster.nodata
        rivmsk = rivwth_org != nodata
        rivwth_org = np.where(rivmsk, rivwth_org, -9999)

        # mask zero and negative riverwidth and reservoirs if present
        mask = rivwth_org > min_wth
        for name in mask_names:
            if name in ds_like:
                mask[ds_like[name].values != ds_like[name].raster.nodata] = False
                logger.debug(f"{name} masked out in rivwth data.")
            else:
                logger.warning(f'mask variable "{name}" not found in maps.')

        # fit predictor
        wth = rivwth_org[mask]
        val = values[mask]
        if fit:
            logger.info(f"Fitting power-law a,b parameters based on {predictor}")
            a, b = _width_fit(wth, val, mask)
        wsim = power_law(val, a, b)
        res = np.abs(wsim - wth)
        outliers = np.logical_and(res > 200, (res / wsim) > 1.0)
        pout = np.sum(outliers) / outliers.size * 100

        # validate
        wsim = power_law(val[~outliers], a, b)
        nse = stats.skills._nse(wsim, wth[~outliers])
        logger.info(
            f'Using "width = a*{predictor}^b"; where a={a:.2f}, b={b:.2f} '
            f"(nse={nse:.3f}; outliers={pout:.2f}%)"
        )

    # compute new riverwidth
    rivmsk = ds_like[rivmsk_name].values != 0
    rivwth_out = np.full(ds_like.raster.shape, -9999.0, dtype=np.float32)
    rivwth_out[rivmsk] = np.maximum(min_wth, power_law(values[rivmsk], a, b))

    # overwrite rivwth data with original data at valid points
    if fill:
        if fill_outliers:
            mask[mask] = ~outliers
            logger.info(f"masking {np.sum(outliers):.0f} outliers")
        rivwth_out[mask] = wth[~outliers]

    fills = "for missing data" if fill else "globally"
    logger.info(f"rivwth set {fills} based on width-{predictor} relationship.")

    # return xarray dataarray
    attrs = dict(_FillValue=-9999, unit="m")
    return xr.DataArray(rivwth_out, dims=ds_like.raster.dims, attrs=attrs)


def _width_fit(
    wth,
    val,
    mask,
    p0=[0.15, 0.65],
):  # rhine uparea based
    outliers = np.full(np.sum(mask), False, dtype=bool)
    a, b = None, None
    # check if sufficient data
    if np.sum(mask) > 10:
        n = 0
        # fit while removing outliers.
        while n < 20:
            if np.sum(mask[mask][~outliers]) > 10:
                pars = curve_fit(
                    f=power_law,
                    xdata=val[~outliers],
                    ydata=wth[~outliers],
                    p0=p0,
                    bounds=([0.01, 0.01], [100.0, 1.0]),
                )[0]
                # print(n, pars, np.sum(~outliers))
                if np.allclose(p0, pars):
                    break
                # outliers based on absolute and relative diff
                wsim = power_law(val, *pars)
                res = np.abs(wsim - wth)
                outliers = np.logical_and(res > 200, (res / wsim) > 1.0)
                if np.sum(outliers) > np.sum(mask) * 0.2:  # max 20 % outliers
                    pars = p0
                    outliers[:] = False
                    logger.warning("Fit not successful.")
                    break
                p0 = pars
                n += 1
        a, b = pars
    elif a is None or b is None:
        a, b = p0
        logger.warning("Insufficient data points, using global parameters.")

    return a, b


def _precip(ds_like, flwdir, da_precip):
    """Do simple precipication method."""
    precip = (
        da_precip.raster.reproject_like(ds_like, method=RESAMPLING["precip"]).values
        / 1e3
    )  # [m/yr]
    precip = np.maximum(precip, 0)
    area = ds_like.raster.area_grid()
    # 10 x average flow
    accu_precip = flwdir.accuflux(precip * area / (86400 * 365) * 10)  # [m3/s]
    return accu_precip


def _discharge(ds_like, flwdir, da_precip, da_climate):
    """Do discharge method."""
    # read clim classes and regression parameters from data dir
    precip_fn = da_precip.name
    climate_fn = da_climate.name
    fn_regr = join(DATADIR, "rivwth", f"regr_{precip_fn}.csv")
    fn_clim = join(DATADIR, "rivwth", f"{climate_fn}.csv")
    clim_map = pd.read_csv(fn_clim, index_col="class")
    regr_map = pd.read_csv(fn_regr, index_col="source").loc[climate_fn]
    regr_map = regr_map.set_index("base_class")

    # get overall catchment climate classification (based on mode)
    # TODO: convert to xarray method by using scipy.stats.mode in xr.apply_ufuncs
    # TODO: reproject climate and mask cells outside basin
    np_climate = da_climate.values
    basin_climate = np.bincount(np_climate.flatten()).argmax()
    # convert detailed climate classification to higher level base class
    base_class, class_name = clim_map.loc[basin_climate]
    logger.debug(f"Basin base climate classification: {base_class} ({class_name})")
    params = regr_map.loc[base_class]

    # precipitation (TO-DO: add checks that were present in old version?)
    precip = da_precip.raster.reproject_like(
        ds_like, method=RESAMPLING["precip"]
    ).values
    # apply scaling factor to make sure accumulated precipitation will stay
    # (approximately) the same
    scaling_factor_1 = np.round(da_precip.raster.res[0] / ds_like.raster.res[0], 4)
    scaling_factor_2 = np.round(da_precip.raster.res[1] / ds_like.raster.res[1], 4)
    scaling_factor = (scaling_factor_1 + scaling_factor_2) / 2
    precip = precip / scaling_factor**2

    # derive cell areas (m2)
    areagrid = ds_like.raster.area_grid() / 1e6

    # calculate "local runoff" (note: set missing in precipitation to zero)
    runoff = (np.maximum(precip, 0) * params["precip"]) + (areagrid * params["area"])
    runoff = np.maximum(runoff, 0)  # make sure runoff is not negative

    # get discharges
    discharge = flwdir.accuflux(runoff, nodata=NODATA["discharge"], direction="up")

    return discharge, (params["discharge_a"], params["discharge_b"])



import platform
from pathlib import Path
from typing import Callable
from unittest.mock import MagicMock, PropertyMock

import numpy as np
import pytest
import xarray as xr
from hydromt import DataCatalog
from hydromt.model import ModelRoot
from pyproj.crs import CRS
from pytest_mock import MockerFixture

from hydromt_wflow import WflowSbmModel
from hydromt_wflow.components import WflowConfigComponent, WflowStaticmapsComponent

SUBDIR = ""
if platform.system().lower() != "windows":
    SUBDIR = "linux64"
TEST_COMPONENT_ROOT_FOLDER = Path(__file__).parent


## OS related fixture
@pytest.fixture(scope="session")
def mount_string() -> str:
    if platform.system().lower() == "windows":
        return "d:/"
    return "/d/"  # Posix paths


## Data directories
@pytest.fixture(scope="session")
def cached_models() -> Path:
    p = Path(TEST_COMPONENT_ROOT_FOLDER, "..", "..", "examples", SUBDIR)
    assert p.is_dir()
    return p.resolve()


@pytest.fixture(scope="session")
def model_subbasin_cached(cached_models: Path) -> Path:
    p = Path(cached_models, "wflow_piave_subbasin")
    assert p.is_dir()
    return p


## Model related fixtures
@pytest.fixture
def mock_model_factory(
    mocker: MockerFixture, tmp_path: Path
) -> Callable[[Path, str], WflowSbmModel]:
    def _factory(path: Path = tmp_path, mode: str = "w") -> WflowSbmModel:
        model = mocker.create_autospec(WflowSbmModel)
        model.root = ModelRoot(path, mode=mode)
        model.data_catalog = mocker.create_autospec(DataCatalog)
        model.crs = CRS.from_epsg(4326)
        model._MAPS = {}
        return model

    return _factory


@pytest.fixture
def mock_model_staticmaps(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    grid_dummy_data: xr.DataArray,
) -> WflowSbmModel:
    # Add a GridComponent to mock model
    mock_model = mock_model_factory()
    staticmaps = WflowStaticmapsComponent(mock_model)
    staticmaps._data = grid_dummy_data.to_dataset(name="basin", promote_attrs=True)

    type(mock_model).components = PropertyMock(
        side_effect=lambda: {"staticmaps": staticmaps}
    )
    type(mock_model).staticmaps = PropertyMock(side_effect=lambda: staticmaps)
    # Mock the get_component method of mock_model to return the staticmaps component
    mock_model.get_component = MagicMock(name="staticmaps", return_value=staticmaps)

    return mock_model


@pytest.fixture
def mock_model_staticmaps_factory(
    grid_dummy_data: xr.DataArray,
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    tmp_path: Path,
) -> WflowSbmModel:
    def factory(path: Path = tmp_path, mode: str = "w"):
        mock_model = mock_model_factory(path, mode)
        staticmaps = WflowStaticmapsComponent(mock_model)
        staticmaps._data = grid_dummy_data.to_dataset(name="basin", promote_attrs=True)
        mock_model.components = {"staticmaps": staticmaps}
        type(mock_model).components = PropertyMock(
            side_effect=lambda: {"staticmaps": staticmaps}
        )
        type(mock_model).staticmaps = PropertyMock(side_effect=lambda: staticmaps)
        mock_model.get_component = MagicMock(name="staticmaps", return_value=staticmaps)
        return mock_model

    return factory


@pytest.fixture
def mock_model_staticmaps_config_factory(
    mock_model_factory: Callable[[Path, str], WflowSbmModel],
    tmp_path: Path,
) -> WflowSbmModel:
    def factory(
        path: Path = tmp_path, mode: str = "w", config_filename: str = "wflow_sbm.toml"
    ):
        mock_model = mock_model_factory(path, mode)
        staticmaps = WflowStaticmapsComponent(mock_model)
        staticmaps._data = None
        config = WflowConfigComponent(mock_model, filename=config_filename)
        config._data = None

        mock_model.components = {"staticmaps": staticmaps, "config": config}
        type(mock_model).components = PropertyMock(
            side_effect=lambda: {"staticmaps": staticmaps, "config": config}
        )
        type(mock_model).staticmaps = PropertyMock(side_effect=lambda: staticmaps)
        type(mock_model).config = PropertyMock(side_effect=lambda: config)
        mock_model.get_component = MagicMock(name="staticmaps", return_value=staticmaps)
        return mock_model

    return factory


## Extra data structures
@pytest.fixture(scope="session")
def config_dummy_data() -> dict:
    data = {
        "biem": "bam",
        "time": {"sometime": "now"},
        "foo": {"bar": "baz", "bip": "bop"},
    }
    return data


@pytest.fixture
def config_dummy_document(tmp_path: Path) -> dict:
    data = {}
    data.update(
        {
            "foo": "bar",
            "baz": {
                "file1": Path(tmp_path, "tmp.txt").as_posix(),
                "file2": "tmp/tmp.txt",
            },
            "spooky": {"ghost": [1, 2, 3]},
        }
    )
    return data


@pytest.fixture
def cyclic_layer() -> xr.DataArray:
    da = xr.DataArray(
        np.ones((12, 2, 2)),
        coords={
            "time": range(1, 13),
            "lat": range(2),
            "lon": range(2),
        },
        dims=["time", "lat", "lon"],
    )
    da.raster.set_nodata(-9999)
    return da


@pytest.fixture
def cyclic_layer_large() -> xr.DataArray:
    da = xr.DataArray(
        np.ones((365, 2, 2)),
        coords={
            "time": range(1, 366),
            "lat": range(2),
            "lon": range(2),
        },
        dims=["time", "lat", "lon"],
    )
    da.raster.set_nodata(-9999)
    return da


@pytest.fixture
def forcing_layer() -> xr.DataArray:
    time = np.arange(
        np.datetime64("2000-01-01"), np.datetime64("2000-01-21"), np.timedelta64(1, "D")
    ).astype("datetime64[ns]")
    da = xr.DataArray(
        np.ones((20, 2, 2)),
        coords={
            "time": time,
            "lat": range(2),
            "lon": range(2),
        },
        dims=["time", "lat", "lon"],
        attrs={
            "crs": "EPSG:4326",
            "grid_mapping_name": "latitude_longitude",
        },
    )
    da.raster.set_nodata(-9999)
    da.name = "foo"
    return da


@pytest.fixture
def forcing_layer_path(tmp_path: Path, forcing_layer: xr.DataArray):
    p = Path(tmp_path, "tmp_forcing.nc")
    forcing_layer.to_netcdf(p)
    assert p.is_file()
    return p


@pytest.fixture(scope="session")
def grid_dummy_data() -> xr.DataArray:
    """Create a dummy grid data array."""
    data = xr.DataArray(
        data=[[1, 2], [3, 4]],
        dims=["y", "x"],
        name="dummy_grid",
        coords={
            "y": [0, 1],
            "x": [0, 1],
        },
        attrs={
            "crs": "EPSG:4326",
            "grid_mapping_name": "latitude_longitude",
        },
    )
    return data


@pytest.fixture
def mask_layer() -> xr.DataArray:
    data = np.ones((2, 2))
    data[1, 1] = -9999
    da = xr.DataArray(
        data,
        coords={
            "lat": range(2),
            "lon": range(2),
        },
        dims=["lat", "lon"],
    )
    da.raster.set_nodata(-9999)
    return da


@pytest.fixture
def static_file(tmp_path: Path, static_layer: xr.DataArray) -> Path:
    p = Path(tmp_path, "tmp.nc")
    ds = static_layer.to_dataset(name="layer1")
    ds["layer2"] = static_layer.where(
        static_layer == static_layer.raster.nodata,
        static_layer * 2,
    )
    ds = ds.raster.gdal_compliant()
    ds.to_netcdf(p)
    return p



"""HydroMT wflow workflows."""

from hydromt_wflow.workflows.basemaps import *
from hydromt_wflow.workflows.connect import *
from hydromt_wflow.workflows.demand import *
from hydromt_wflow.workflows.forcing import *
from hydromt_wflow.workflows.gauges import *
from hydromt_wflow.workflows.glaciers import *
from hydromt_wflow.workflows.landuse import *
from hydromt_wflow.workflows.reservoirs import *
from hydromt_wflow.workflows.river import *
from hydromt_wflow.workflows.rootzoneclim import *
from hydromt_wflow.workflows.soilgrids import *
from hydromt_wflow.workflows.soilparams import *
from hydromt_wflow.workflows.states import *



import os
from typing import TYPE_CHECKING

from hydromt.model.components import TablesComponent
from hydromt.model.steps import hydromt_step

if TYPE_CHECKING:
    from hydromt.model.model import Model


class WflowTablesComponent(TablesComponent):
    """Wflow specific tables component."""

    def __init__(self, model: "Model", filename: str = "{name}.csv"):
        super().__init__(model, filename=filename)

    @hydromt_step
    def read(self, filename: str | None = None, **kwargs) -> None:
        """Read tables at provided or default file path if none is provided."""
        if filename is None:
            filename = self._generate_filename_from_staticmaps()
        super().read(filename=filename, **kwargs)

    @hydromt_step
    def write(self, filename: str | None = None, **kwargs) -> None:
        """Write tables at provided or default file path if none is provided."""
        if filename is None:
            filename = self._generate_filename_from_staticmaps()
        super().write(filename=filename, **kwargs)

    def _generate_filename_from_staticmaps(self) -> str | None:
        static_maps_path = self.model.config.get_value("input.path_static")
        if static_maps_path is None:
            return None

        static_maps_folder = os.path.dirname(static_maps_path)
        if not static_maps_folder:
            static_maps_folder = "."

        filename = f"{static_maps_folder}/{{name}}.csv"
        return filename



"""HydroMT plugin for wflow models."""

from hydromt_wflow.version import __version__
from hydromt_wflow.wflow_base import WflowBaseModel
from hydromt_wflow.wflow_sbm import WflowSbmModel
from hydromt_wflow.wflow_sediment import WflowSedimentModel

__all__ = ["WflowBaseModel", "WflowSbmModel", "WflowSedimentModel"]



"""Workflow for wflow model states."""

import numpy as np
import pandas as pd
import xarray as xr
from hydromt.gis import full_like
from hydromt.model.processes.grid import grid_from_constant

from hydromt_wflow.utils import get_grid_from_config

__all__ = ["prepare_cold_states"]


def prepare_cold_states(
    ds_like: xr.Dataset,
    config: dict,
    timestamp: str = None,
    mask_name_land: str = "subcatchment",
    mask_name_river: str = "river_mask",
) -> tuple[xr.Dataset, dict[str, str]]:
    """
    Prepare cold states for Wflow.

    Compute cold states variables:

    * **soil_saturated_depth**: saturated store [mm]
    * **snow_leq_depth**: snow storage [mm]
    * **soil_temp**: top soil temperature [°C]
    * **soil_unsaturated_depth**: amount of water in the unsaturated store, per layer
      [mm]
    * **snow_water_depth**: liquid water content in the snow pack [mm]
    * **vegetation_water_depth**: canopy storage [mm]
    * **river_instantaneous_q**: river discharge [m3/s]
    * **river_h**: river water level [m]
    * **subsurface_q**: subsurface flow [m3/d]
    * **land_h**: land water level [m]
    * **land_instantaneous_q** or **land_instantaneous_qx**+**land_instantaneous_qy**:
      overland flow for kinwave [m3/s] or overland flow in x/y directions for
      local_inertial [m3/s]

    If reservoirs, also adds:

    * **reservoir_water_level**: reservoir water level [m]

    If glaciers, also adds:

    * **glacier_leq_depth**: water within the glacier [mm]

    If paddy, also adds:

    * **demand_paddy_h**: water on the paddy fields [mm]

    Parameters
    ----------
    ds_like : xr.Dataset
        Dataset containing the staticmaps grid and variables to prepare some of the
        states.

        * Required variables: `mask_name_land`, `mask_name_river`

        * Other required variables (exact name will be read from the wflow config):
            soil_brooks_corey_c, soilthickness,
            theta_s, theta_r, ksat_vertical, f, slope, subsurface_ksat_horizontal_ratio

        * Optional variables (exact name from the wflow config):
            glacierstore, reservoir_water_surface__initial_elevation
    config : dict
        Wflow configuration dictionary.
    timestamp : str, optional
        Timestamp of the cold states. By default uses the starttime
        from the config.
    mask_name_land : str, optional
        Name of the land mask variable in the ds_like dataset. By default
        subcatchment.
    mask_name_river : str, optional
        Name of the river mask variable in the ds_like dataset. By default
        river_mask.

    Returns
    -------
    xr.Dataset
        Dataset containing the cold states.
    dict
        Config dictionary with the cold states variable names.
    """
    # Defaults
    nodata = -9999.0
    dtype = "float32"

    # Times
    if timestamp is None:
        # states time = starttime for Wflow.jl > 0.7.0)
        if "time" in config:
            starttime = pd.to_datetime(config["time"].get("starttime", None))
            timestamp = starttime  # - pd.Timedelta(seconds=timestepsecs)
        if timestamp is None:
            raise ValueError("No timestamp provided and no starttime in config.")
    else:
        timestamp = pd.to_datetime(timestamp)
    timestepsecs = config.get("time.timestepsecs", 86400)

    # Create empty dataset
    ds_out = xr.Dataset()

    # Base output config dict for states
    states_config = {
        "state.variables.soil_water_saturated_zone__depth": "soil_saturated_depth",
        "state.variables.snowpack_dry_snow__leq_depth": "snow_leq_depth",
        "state.variables.soil_surface__temperature": "soil_temp",
        "state.variables.soil_layer_water_unsaturated_zone__depth": "soil_unsaturated_depth",  # noqa: E501
        "state.variables.snowpack_liquid_water__depth": "snow_water_depth",
        "state.variables.vegetation_canopy_water__depth": "vegetation_water_depth",
        "state.variables.subsurface_water__volume_flow_rate": "subsurface_q",
        "state.variables.land_surface_water__depth": "land_h",
        "state.variables.river_water__instantaneous_volume_flow_rate": "river_instantaneous_q",  # noqa: E501
        "state.variables.river_water__depth": "river_h",
    }

    # Map with constant values or zeros for basin
    zeromap = [
        "soil_temp",
        "snow_leq_depth",
        "snow_water_depth",
        "vegetation_water_depth",
        "land_h",
    ]
    land_routing = config["model"].get("land_routing", "kinematic_wave")
    if land_routing == "local_inertial":
        zeromap.extend(["land_instantaneous_qx", "land_instantaneous_qy"])
        states_config[
            "state.variables.land_surface_water__x_component_of_instantaneous_volume_flow_rate"
        ] = "land_instantaneous_qx"
        states_config[
            "state.variables.land_surface_water__y_component_of_instantaneous_volume_flow_rate"
        ] = "land_instantaneous_qy"
    else:
        zeromap.extend(["land_instantaneous_q"])
        states_config[
            "state.variables.land_surface_water__instantaneous_volume_flow_rate"
        ] = "land_instantaneous_q"

    for var in zeromap:
        if var == "soil_temp":
            value = 10.0
        else:
            value = 0.0
        da_param = grid_from_constant(
            ds_like,
            constant=value,
            name=var,
            dtype=dtype,
            nodata=nodata,
            mask_name=mask_name_land,
        )
        ds_out[var] = da_param

    # soil_unsaturated_depth (zero per layer)
    # layers are based on brooks_corey_c parameter
    c = get_grid_from_config(
        "soil_layer_water__brooks_corey_exponent",
        config=config,
        grid=ds_like,
    )
    usld = full_like(c, nodata=nodata)
    for sl in usld["layer"]:
        usld.loc[dict(layer=sl)] = xr.where(ds_like[mask_name_land], 0.0, nodata)
    ds_out["soil_unsaturated_depth"] = usld

    # Compute other soil states
    # Get required variables from config
    st = get_grid_from_config(
        "soil__thickness",
        config=config,
        grid=ds_like,
    )
    ts = get_grid_from_config(
        "soil_water__saturated_volume_fraction",
        config=config,
        grid=ds_like,
    )
    tr = get_grid_from_config(
        "soil_water__residual_volume_fraction",
        config=config,
        grid=ds_like,
    )
    ksh = get_grid_from_config(
        "subsurface_water__horizontal_to_vertical_saturated_hydraulic_conductivity_ratio",
        config=config,
        grid=ds_like,
    )
    ksv = get_grid_from_config(
        "soil_surface_water__vertical_saturated_hydraulic_conductivity",
        config=config,
        grid=ds_like,
    )
    f = get_grid_from_config(
        "soil_water__vertical_saturated_hydraulic_conductivity_scale_parameter",
        config=config,
        grid=ds_like,
    )
    sl = get_grid_from_config(
        "land_surface__slope",
        config=config,
        grid=ds_like,
    )

    # soil_saturated_depth
    swd = 0.85 * st * (ts - tr)
    swd = swd.where(ds_like[mask_name_land] > 0, nodata)
    swd.raster.set_nodata(nodata)
    ds_out["soil_saturated_depth"] = swd

    # subsurface_q
    zi = np.maximum(0.0, st - swd / (ts - tr))
    kh0 = ksh * ksv * 0.001 * (86400 / timestepsecs)
    ssf = (kh0 * np.maximum(0.00001, sl) / (f * 1000)) * (
        np.exp(-f * 1000 * zi * 0.001)
    ) - (np.exp(-f * 1000 * st) * np.sqrt(ds_like.raster.area_grid()))
    ssf = ssf.where(ds_like[mask_name_land] > 0, nodata)
    ssf.raster.set_nodata(nodata)
    ds_out["subsurface_q"] = ssf

    # River
    zeromap_riv = ["river_instantaneous_q", "river_h"]
    # 1D floodplain
    if config["model"].get("floodplain_1d__flag", False):
        zeromap_riv.extend(["floodplain_instantaneous_q", "floodplain_h"])
        states_config[
            "state.variables.floodplain_water__instantaneous_volume_flow_rate"
        ] = "floodplain_instantaneous_q"
        states_config["state.variables.floodplain_water__depth"] = "floodplain_h"
    for var in zeromap_riv:
        value = 0.0
        da_param = grid_from_constant(
            ds_like,
            constant=value,
            name=var,
            dtype=dtype,
            nodata=nodata,
            mask_name=mask_name_river,
        )
        da_param = da_param.rename(var)
        ds_out[var] = da_param

    # reservoir
    if config["model"].get("reservoir__flag", False):
        rl = get_grid_from_config(
            "reservoir_water_surface__initial_elevation",
            config=config,
            grid=ds_like,
        )
        rl = rl.where(rl != rl.raster.nodata, nodata)
        rl.raster.set_nodata(nodata)
        ds_out["reservoir_water_level"] = rl

        states_config["state.variables.reservoir_water_surface__elevation"] = (
            "reservoir_water_level"
        )

    # glacier
    if config["model"].get("glacier__flag", False):
        gs_vn = get_grid_from_config(
            "glacier_ice__initial_leq_depth",
            config=config,
            grid=ds_like,
        )
        if gs_vn.name in ds_like:
            ds_out["glacier_leq_depth"] = ds_like[gs_vn.name]
        else:
            glacstore = grid_from_constant(
                ds_like,
                value=5500.0,
                name="glacier_leq_depth",
                nodata=nodata,
                dtype=dtype,
                mask=mask_name_land,
            )
            ds_out["glacier_leq_depth"] = glacstore

        states_config["state.variables.glacier_ice__leq_depth"] = "glacier_leq_depth"

    # paddy
    if config["model"].get("water_demand.paddy__flag", False):
        h_paddy = grid_from_constant(
            ds_like,
            value=0.0,
            name="demand_paddy_h",
            nodata=nodata,
            dtype=dtype,
            mask=mask_name_land,
        )
        ds_out["demand_paddy_h"] = h_paddy

        states_config["state.variables.paddy_surface_water__depth"] = "demand_paddy_h"

    # Add time dimension
    ds_out = ds_out.expand_dims(dim=dict(time=[timestamp]))

    return ds_out, states_config



"""Custom wflow config component module."""

import logging
from pathlib import Path
from typing import Any

from hydromt import hydromt_step
from hydromt.model import Model
from hydromt.model.components import ConfigComponent
from hydromt.readers import read_toml
from hydromt.writers import write_toml

from hydromt_wflow import utils
from hydromt_wflow.components.utils import make_config_paths_relative

__all__ = ["WflowConfigComponent"]

logger = logging.getLogger(f"hydromt.{__name__}")


class WflowConfigComponent(ConfigComponent):
    """Manage the wflow TOML configuration file for model simulations/settings.

    ``WflowConfigComponent`` data is stored in a dictionary. The component
    is used to prepare and update model simulations/settings of the wflow model.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "wflow_sbm.toml",
        default_template_filename: str | None = None,
    ):
        """Manage configuration files for model simulations/settings.

        Parameters
        ----------
        model : Model
            HydroMT model instance
        filename : str
            A path relative to the root where the configuration file will
            be read and written if user does not provide a path themselves.
            By default 'wflow_sbm.toml'
        default_template_filename : str, optional
            A path to a template file that will be used as default in the ``create``
            method to initialize the configuration file if the user does not provide
            their own template file. This can be used by model plugins to provide a
            default configuration template. By default None.
        """
        super().__init__(
            model,
            filename=filename,
            default_template_filename=default_template_filename,
        )

    def _initialize(self, skip_read=False) -> None:
        """Initialize the model config."""
        if self._data is None:
            self._data = {}
            if not skip_read:
                # no check for read mode here
                # model config is read if in read-mode and it exists
                # default config if in write-mode
                self.read()

    @property
    def data(self) -> dict:
        """Model config values."""
        if self._data is None:
            self._initialize()
        return self._data

    @hydromt_step
    def read(
        self,
        filename: str | None = None,
    ):
        """
        Read the wflow configuration file from <root/filename>.

        If filename is not provided, will default to <root>/{self._filename} or default
        template configuration file if the model is in write only mode (ie build).

        If filename is provided, it will check if the path is absolute and else will
        assume the given path is relative to the model root.

        """
        self._initialize(skip_read=True)

        # Check if user-defined path or template should be used
        if not filename:
            # Write only mode > read default config
            if (
                not self.root.is_reading_mode()
                and self._default_template_filename is not None
            ):
                prefix = "default"
                read_path = Path(self._default_template_filename)
            else:
                prefix = "model"
                read_path = Path(self.root.path, self._filename)
        else:
            prefix = "user defined"
            # Check if user-defined file is absolute (ie file exists)
            if Path(filename).is_file():
                read_path = Path(filename)
            else:
                read_path = Path(self.root.path, filename)

        # Check if the file exists
        if read_path.is_file():
            logger.info(f"Reading {prefix} config file from {read_path.as_posix()}.")
        else:
            logger.warning(
                f"No config was found at {read_path.as_posix()}. "
                "It wil be initialized as empty dict"
            )
            return

        # Read the data and set it in the document
        self._data = read_toml(read_path)

    @hydromt_step
    def write(
        self,
        filename: str | None = None,
        config_root: Path | str | None = None,
    ):
        """
        Write the configuration to a file.

        The file is written to ``<root>/<filename>`` by default, or to
        ``<config_root>/<filename>`` if a ``config_root`` is provided.

        Parameters
        ----------
        filename : str, optional
            Name of the config file. By default None to use the default name
            self._filename.
        config_root : str, optional
            Root folder to write the config file if different from model root (default).
            Can be absolute or relative to model root.
        """
        # Check for config_root
        path = filename or self._filename
        if config_root is not None:
            path = Path(config_root, path)

        self.root._assert_write_mode()
        # If there is data
        if self.data:
            p = path or self._filename

            # Sort the path
            write_path = Path(self.root.path, p)
            logger.info(f"Writing model config to {write_path.as_posix()}.")
            write_path.parent.mkdir(parents=True, exist_ok=True)

            # Solve the path in the data
            # Extra check for dir_input
            rel_path = Path(write_path.parent, self.get_value("dir_input", fallback=""))
            write_data = make_config_paths_relative(self.data, rel_path)
            write_toml(write_path, write_data)
        else:
            logger.warning("Model config has no data, skip writing.")

    def get_value(
        self,
        key: str,
        fallback: Any | None = None,
        abs_path: bool = False,
    ) -> Any | None:
        """Get config options.

        Parameters
        ----------
        key : str
            Keys are a string with '.' indicating a new level: ('key1.key2')
        fallback : Any, optional
            Fallback value if key(s) not found in config, by default None.
        abs_path: bool, optional
            If True return the absolute path relative to the model root,
            by default False.
        """
        # Refer to utils function of get_config
        return utils.get_config(
            self.data,
            key,
            root=self.root.path,
            fallback=fallback,
            abs_path=abs_path,
        )

    def remove(self, *args: str, errors: str = "raise") -> Any:
        """
        Remove a config key and return its value.

        Parameters
        ----------
        key: str, tuple[str, ...]
            Key to remove from the config.
            Can be a dotted toml string when providing a list of strings.
        errors: str, optional
            What to do if the key is not found. Can be "raise" (default) or "ignore".

        Returns
        -------
        The popped value, or raises a KeyError if the key is not found.
        """
        args = list(args)
        if len(args) == 1 and "." in args[0]:
            args = args[0].split(".") + args[1:]

        current = self.data
        for index, key in enumerate(args):
            if current is None:
                if errors == "ignore":
                    return None
                else:
                    raise KeyError(f"Key {'.'.join(args)} not found in config.")

            if index == len(args) - 1:
                # Last key, pop it
                if errors == "ignore":
                    current = current.pop(key, None)
                else:
                    current = current.pop(key)
                break

            # Not the last key, go deeper
            current = current.get(key)
        return current

    def remove_reservoirs(
        self, input: list[str | None] = [], state: list[str | None] = []
    ):
        """Remove all reservoir related config options."""
        # a. change reservoir__flag = true to false
        self.set("model.reservoir__flag", False)
        # b. remove reservoir state
        for state_var in state:
            if state_var is not None:
                self.remove(f"state.variables.{state_var}", errors="ignore")
        # c. remove reservoir input
        for input_var in input:
            if input_var is not None:
                # find if variable in input/static/cyclic/forcing
                input_var = utils.get_wflow_var_fullname(input_var, self.data)
                self.remove(input_var, errors="ignore")



"""Utility of the wflow components."""

import re
from os.path import relpath
from pathlib import Path
from typing import Any

import numpy as np
import xarray as xr

MOUNT_PATTERN = re.compile(r"(^\/(\w+)\/|^(\w+):\/).*$")


## Config/ pathing related
def _mount(
    value: str,
) -> str | None:
    """Get the mount of a path.

    As this not properly solved currently by pathlib or os.
    """
    m = MOUNT_PATTERN.match(value)
    if m is None:
        return None
    return m.group(1)


def _relpath(
    value: Any,
    root: Path,
) -> str | Any:
    """Generate a relative path.

    Being able to go either up or down in directories.
    Also return the original path if the mount differs.
    Otherwise it will error..
    """
    if not isinstance(value, (Path, str)) or not Path(value).is_absolute():
        return value
    value = Path(value)
    try:
        if _mount(value.as_posix()) == _mount(root.as_posix()):
            value = Path(relpath(value, root))
    except ValueError:
        pass  # `value` path is not relative to root
    return value.as_posix()


def make_config_paths_relative(
    data: dict,
    root: Path,
):
    """Make the configurations path relative to the root.

    This only concerns itself with paths that are absolute and on
    the same mount.

    Parameters
    ----------
    data : dict
        The configurations in a dictionary format.
    root : Path
        The root to which the paths are made relative.
        Most of the time, this will be the parent directory of the
        configurations file.
    """
    for key, val in data.items():
        if isinstance(val, dict):
            data.update({key: make_config_paths_relative(val, root)})
        else:
            data.update({key: _relpath(val, root)})
    return data


## Grid related
def get_mask_layer(mask: str | xr.DataArray | None, *args) -> xr.DataArray | None:
    """Get the proper mask layer based on itself or a layer in a Dataset.

    Parameters
    ----------
    mask : str | xr.DataArray | None
        The mask itself or the name of the mask layer in another dataset.
    *args : list
        These have to be xarray Datasets in which the mask (as a string)
        can be present
    """
    if mask is None:
        return None
    if isinstance(mask, xr.DataArray):
        return mask != mask.raster.nodata
    if not isinstance(mask, str):
        raise ValueError(f"Unknown type for determining mask: {type(mask).__name__}")
    for ds in args:
        if mask in ds:
            return ds[mask] != ds[mask].raster.nodata
    return None  # Nothin found


def align_grid_with_region(
    data: xr.DataArray,
    region_grid: xr.DataArray,
    region_component_name: str,
) -> xr.DataArray:
    """Align the data grid with the region grid.

    Aligns the name of the x and y dimensions and flips the y dimension if needed.

    Parameters
    ----------
    data : xr.DataArray
        The data array to align.
    region_grid : xr.DataArray
        The region grid to align with.
    region_component_name : str
        The name of the region component.

    Returns
    -------
    xr.DataArray
        The aligned data array.

    Raises
    ------
    ValueError
        If the data grid is not identical to the region grid.
    """
    if not data.raster.identical_grid(region_grid):
        y_dim = region_grid.raster.y_dim
        x_dim = region_grid.raster.x_dim
        # First try to rename dimensions
        data = data.rename(
            {
                data.raster.x_dim: x_dim,
                data.raster.y_dim: y_dim,
            }
        )
        # Flip latitude if needed
        if (
            np.diff(data[y_dim].values)[0] > 0
            and np.diff(region_grid[y_dim].values)[0] < 0
        ):
            data = data.reindex({y_dim: data[y_dim][::-1]})

    if not data.raster.identical_grid(region_grid):
        raise ValueError(
            f"Data grid must be identical to {region_component_name} component"
        )

    return data


def test_equal_grid_data(
    grid: xr.Dataset, other_grid: xr.Dataset
) -> tuple[bool, dict[str, str]]:
    """
    Test if two grid datasets are equal.

    Checks the CRS, dimensions, and data variables for equality.

    Parameters
    ----------
    grid : xr.Dataset
        The first grid dataset to compare.
    other_grid : xr.Dataset
        The second grid dataset to compare.

    Returns
    -------
    tuple[bool, dict[str, str]]
        True if the grids are equal, and a dict with the associated errors per property
        checked.
    """
    errors: dict[str, str] = {}
    invalid_maps: dict[str, str] = {}
    invalid_coords: dict[str, str] = {}

    # Check if grid is empty
    if len(grid) == 0:
        if len(other_grid) == 0:
            return True, {}
        else:
            errors["grid"] = "first grid is empty, second is not"
            return False, errors

    # Check CRS and dims
    maps = grid.raster.vars

    if not np.all(grid.raster.crs == other_grid.raster.crs):
        errors["crs"] = "the two grids have different crs"

    # check on dims names and values
    for dim in other_grid.dims:
        try:
            xr.testing.assert_identical(other_grid[dim], grid[dim])
        except AssertionError:
            errors["dims"] = f"dim {dim} not identical"
        except KeyError:
            errors["dims"] = f"dim {dim} not in grid"

    # Check if new maps in other grid
    new_maps = []
    for name in other_grid.raster.vars:
        if name not in maps:
            new_maps.append(name)
    if len(new_maps) > 0:
        errors["Other grid has additional maps"] = f"{', '.join(new_maps)}"

    # Check per map (dtype, value, nodata)
    missing_maps = []
    for name in maps:
        map0 = grid[name].fillna(0)
        if name not in other_grid.data_vars:
            missing_maps.append(name)
            continue
        map1 = other_grid[name].fillna(0)

        # hilariously np.nan == np.nan returns False, hence the additional check
        equal_nodata = map0.raster.nodata == map1.raster.nodata
        if not equal_nodata and (
            np.isnan(map0.raster.nodata) and np.isnan(map1.raster.nodata)
        ):
            equal_nodata = True

        if (
            not np.allclose(map0, map1, atol=1e-3, rtol=1e-3)
            or map0.dtype != map1.dtype
            or not equal_nodata
        ):
            if len(map0.dims) > 2:  # 3 dim map
                map0 = map0[0, :, :]
                map1 = map1[0, :, :]
            # Check on dtypes
            err = (
                ""
                if map0.dtype == map1.dtype
                else f"{map1.dtype} instead of {map0.dtype}"
            )
            # Check on nodata
            err = (
                err
                if equal_nodata
                else f"nodata {map1.raster.nodata} instead of \
{map0.raster.nodata}; {err}"
            )
            not_close = ~np.equal(map0, map1)
            n_cells = int(np.sum(not_close))
            if n_cells > 0:
                diff = (map0.values - map1.values)[not_close].mean()
                err = f"mean diff ({n_cells:d} cells): {diff:.4f}; {err}"
            invalid_maps[name] = err
    for coord in grid.coords:
        if coord in other_grid.coords:
            try:
                xr.testing.assert_identical(other_grid[coord], grid[coord])
            except AssertionError:
                invalid_coords[coord] = "not identical"
            except KeyError:
                invalid_coords[coord] = "not in grid"

    if len(invalid_coords) > 0:
        errors[f"{len(invalid_coords)} invalid coords"] = invalid_coords

    if len(missing_maps) > 0:
        errors["Other grid is missing maps"] = f"{', '.join(missing_maps)}"

    if len(invalid_maps) > 0:
        errors[f"{len(invalid_maps)} invalid maps"] = invalid_maps

    return len(errors) == 0, errors



"""Functions for individual soil parameters."""

import numpy as np
import xarray as xr
from hydromt.gis import raster_utils

__all__ = [
    "ksat_horizontal_ratio",
    "ksatver_vegetation",
]


def ksat_horizontal_ratio(
    da: xr.DataArray,
    ds_like: xr.Dataset,
    resampling_method: str,
) -> xr.DataArray:
    """Create KsatHorfrac map.

    Based on the data properties of the WflowSbmModel.

    Parameters
    ----------
    da : xr.DataArray
        KsathorFrac values, i.e. an xarray.DataArray containing the values from \
the predefined KsatHorFrac map.
    ds_like : xr.Dataset
        Dataset at model resolution.
    resampling_method : str
        Scale method when up- or downscaling.

    Returns
    -------
    xr.DataArray
        A xarray DataArray containing the scaled KsatHorFrac values.
    """
    # Transfer data to a logarithmic scale
    da = np.log10(da)
    # Reproject the data
    da = da.raster.reproject_like(
        ds_like,
        method=resampling_method,
    )
    # Scale the data back to normal values
    da = np.power(10, da)

    # Fill all nodata holes in the map
    da = da.interpolate_na(dim=da.raster.x_dim, method="linear")
    da = da.interpolate_na(
        dim=da.raster.x_dim, method="nearest", fill_value="extrapolate"
    )

    # Set the default no fill value for doubles
    da = da.fillna(-9999.0)
    da.raster.set_nodata(-9999.0)
    return da


def calc_kv_at_depth(depth, kv_0, f):
    """
    Calculate the kv value at a certain depth.

    Value is based on the kv at the surface, the f parameter that describes the
    exponential decline and the depth.

    Parameters
    ----------
    depth
        Depth at which kv needs to be calculated
    kv_0
        The vertical conductivity at the surface
    f
        The value describing the exponential decline

    Returns
    -------
    kv_z
        The kv value at the requested depth
    """
    exp_arg = -f * depth
    # Clip exponent to float64-safe range
    exp_arg = np.clip(exp_arg, -700, 700)
    kv_z = kv_0 * np.exp(exp_arg)

    return kv_z


def calc_kvfrac(kv_depth, target):
    """Calculate the kvfrac.

    Based on the kv value at a certain depth and the target value.

    Parameters
    ----------
    kv_depth:
        Value of kv at a certain depth
    target:
        Target kv value

    Returns
    -------
    kvfrac:
        The value which kv_depths needs to be multiplied with to reach the target value
    """
    kvfrac = target / kv_depth
    return kvfrac


def update_kvfrac(
    ds_model, kv0_mask, f_mask, wflow_thicknesslayers, target_conductivity
):
    """
    Calculate kvfrac values for each layer.

    Done such that the bottom of the layer equals to the target_conductivity.
    Calculation assumes exponentially declining vertical conductivities, based on the f
    parameter. If no target_conductivity is specified, kvfrac is set to be equal to 1.

    Parameters
    ----------
    ds_model
        Dataset of the wflow model
    kv0_mask
        Values of vertical conductivity at the surface, masked to paddy locations
    f_mask
        Values of the f parameter, masked to paddy locations
    wflow_thicknesslayers
        List of requested layers in the wflow model
    target_conductivity
        List of target conductivities for each layer (None if no target value is
        requested)

    Returns
    -------
    da_kvfrac
        Maps for each layer with the required kvfrac value
    """
    # Convert to np.array
    wflow_thicknesslayers = np.array(wflow_thicknesslayers)
    target_conductivity = np.array(target_conductivity)

    # Prepare empty dataarray
    da_kvfrac = raster_utils.full_like(ds_model["soil_brooks_corey_c"])
    # Set all values to 1
    da_kvfrac = da_kvfrac.where(ds_model["elevtn"].raster.mask_nodata().isnull(), 1.0)

    # Get the actual depths
    wflow_depths = np.cumsum(wflow_thicknesslayers)
    # Find the index of the layers where a kvfrac should be set
    idx = target_conductivity.nonzero()[0]

    # Loop through the target_conductivity values
    for idx, target in enumerate(target_conductivity):
        if target is not None:
            depth = wflow_depths[idx]
            # Calculate the kv at that depth (only for the pixels that have paddy
            # fields)
            kv_depth = calc_kv_at_depth(depth=depth, kv_0=kv0_mask, f=f_mask)
            paddy_values = calc_kvfrac(kv_depth=kv_depth, target=target)
            # Set the values in the correct places
            kvfrac = xr.where(
                paddy_values.raster.mask_nodata().isnull(),
                da_kvfrac.loc[dict(layer=idx)],
                paddy_values,
            )
            kvfrac = kvfrac.fillna(-9999)
            kvfrac.raster.set_nodata(-9999)
            # Update layer in dataarray
            da_kvfrac.loc[dict(layer=idx)] = kvfrac

    return da_kvfrac


def get_ks_veg(ksat_vertical, sndppt, LAI, alfa=4.5, beta=5):
    """
    Based on Bonetti et al. (2021) https://www.nature.com/articles/s43247-021-00180-0.

    Parameters
    ----------
    ksat_vertical  : [xr.DataSet, float]
        saturated hydraulic conductivity from PTF based on soil properties [cm/d].
    sndppt : [xr.DataSet, float]
        percentage sand [%].
    LAI : [xr.DataSet, float]
        Mean annual leaf area index [-].
    alfa : float, optional
        Shape parameter. The default is 4.5 when using LAI.
    beta : float, optional
        Shape parameter. The default is 5 when using LAI.

    Returns
    -------
    ks : [xr.DataSet, float]
        saturated hydraulic conductivity based on soil and vegetation [cm/d].

    """
    # get the saturated hydraulic conductivity with fully developed vegetation.
    ksmax = 10 ** (
        3.5 - 1.5 * sndppt**0.13 + np.log10(ksat_vertical.where(ksat_vertical > 0))
    )
    # get the saturated hydraulic conductivity based on soil and
    # vegetation mean LAI
    ks = ksmax - (ksmax - ksat_vertical) / (1 + (LAI / alfa) ** beta)
    return ks


def ksatver_vegetation(
    ds_like: xr.Dataset,
    sndppt: xr.Dataset,
    alfa: float = 4.5,
    beta: float = 5,
) -> xr.DataArray:
    """
    Calculate saturated hydraulic conductivity based on soil and vegetation [mm/d].

    based on Bonetti et al. (2021) https://www.nature.com/articles/s43247-021-00180-0.

    Parameters
    ----------
    ds_like : xr.Dataset
        Dataset at model resolution.
        The required variables in ds_like are LAI [-],
        ksat_vertical [mm/d] and subcatchment
    sndppt : [xr.DataSet, float]
        percentage sand [%].
    alfa : float, optional
        Shape parameter. The default is 4.5 when using LAI.
    beta : float, optional
        Shape parameter. The default is 5 when using LAI.

    Returns
    -------
    ks : [xr.DataSet, float]
        saturated hydraulic conductivity based on soil and vegetation [mm/d].

    """
    sndppt = sndppt.where(sndppt != sndppt._FillValue, np.nan)
    # reproject to model resolution
    sndppt = sndppt.raster.reproject_like(ds_like, method="average")
    sndppt.raster.set_nodata(np.nan)
    # interpolate to fill missing values
    sndppt = sndppt.raster.interpolate_na("rio_idw")
    # mask outside basin
    sndppt = sndppt.where(ds_like["basins"] > 0)

    # mean annual lai is required (see fig 1 in Bonetti et al. 2021)
    LAI_mean = ds_like["LAI"].mean("time")
    LAI_mean.raster.set_nodata(255.0)

    # in this function, Ksatver should be provided in cm/d
    ksatver_vegetation = get_ks_veg(
        ds_like["ksat_vertical"] / 10, sndppt, LAI_mean, alfa, beta
    )

    # convert back from cm/d to mm/d
    ksatver_vegetation = ksatver_vegetation * 10
    return ksatver_vegetation



"""Tests for the utils module."""

import logging
from os.path import abspath, dirname, join
from pathlib import Path

import numpy as np

from hydromt_wflow import WflowSbmModel, WflowSedimentModel
from hydromt_wflow.utils import get_grid_from_config

TESTDATADIR = Path(dirname(abspath(__file__)), "data")
EXAMPLEDIR = Path(dirname(abspath(__file__)), "..", "examples", "data")


def test_grid_from_config(demda):
    # Create a couple of variables in grid
    grid = demda.to_dataset(name="dem")
    grid["slope"] = demda * 0.1
    grid["mask"] = demda > 0

    # Create config with all options
    config = {
        "input": {
            "dem": "dem",
            "static": {
                "slope": "slope",
                "altitude": {
                    "netcdf_variable_name": "slope",
                    "scale": 10,
                },
            },
            "cyclic": {
                "subsurface_ksat_horizontal_ratio": {"value": 500},
                "ksathorfrac2": {
                    "netcdf_variable_name": "dem",
                    "scale": 0,
                    "offset": 500,
                },
            },
        },
    }

    # Tests
    dem = get_grid_from_config("dem", config=config, grid=grid)
    assert dem.equals(demda)

    slope = get_grid_from_config("slope", config=config, grid=grid)
    assert slope.equals(grid["slope"])

    altitude = get_grid_from_config("altitude", config=config, grid=grid)
    assert altitude.equals(grid["slope"] * 10)

    subsurface_ksat_horizontal_ratio = get_grid_from_config(
        "subsurface_ksat_horizontal_ratio",
        config=config,
        grid=grid,
    )
    assert np.unique(subsurface_ksat_horizontal_ratio.raster.mask_nodata()) == [500]

    ksathorfrac2 = get_grid_from_config(
        "ksathorfrac2",
        config=config,
        grid=grid,
    )
    assert ksathorfrac2.equals(subsurface_ksat_horizontal_ratio)


def test_convert_to_wflow_v1_sbm(tmpdir, caplog):
    # Initialize wflow model
    root = join(EXAMPLEDIR, "wflow_upgrade", "sbm")
    config_fn = "wflow_sbm_v0x.toml"

    wflow = WflowSbmModel(root, config_filename=config_fn, mode="r")

    # Convert to v1
    wflow.upgrade_to_v1_wflow()

    # Check with a test config
    config_fn_v1 = join(TESTDATADIR, "wflow_v0x", "sbm", "wflow_sbm_v1.toml")
    wflow_v1 = WflowSbmModel(root, config_filename=config_fn_v1, mode="r")

    # Set kinematic_wave__adaptive_time_step_flag to false to mirror settings in wflow
    #  v1 config
    wflow.config.data["model"]["kinematic_wave__adaptive_time_step_flag"] = False
    assert wflow.config.test_equal(wflow_v1.config)[0]

    # Checks on extra data in staticmaps
    res_ids = np.unique(
        wflow.staticmaps.data["reservoir_outlet_id"].raster.mask_nodata()
    )
    assert np.all(np.isin([3349.0, 3367.0, 169986.0], res_ids))
    assert np.all(
        np.isin(
            [3.0, 4.0],
            wflow.staticmaps.data["reservoir_rating_curve"].raster.mask_nodata(),
        )
    )

    # Test reinit flag set to False
    wflow = WflowSbmModel(root, config_filename=config_fn, mode="r")
    wflow.config.set("model.reinit", False)
    caplog.set_level(logging.WARNING)
    wflow.upgrade_to_v1_wflow()
    assert "Converting states is not supported by this conversion code" in caplog.text
    assert wflow.config.get_value("model.cold_start__flag") is True


def test_convert_to_wflow_v1_sbm_with_exceptions():
    # Initialize wflow model
    root = join(EXAMPLEDIR, "wflow_upgrade", "sbm")
    config_fn = "wflow_sbm_v0x.toml"

    wflow = WflowSbmModel(root, config_filename=config_fn, mode="r")
    theta_s = wflow.config.remove("input.vertical.theta_s")
    theta_r = wflow.config.remove("input.vertical.theta_r")
    g_ttm = wflow.config.remove("input.vertical.g_ttm")
    kv = wflow.config.remove("input.vertical.kv_0")

    wflow.config.set("input.vertical.θₛ", theta_s)
    wflow.config.set("input.vertical.θᵣ", theta_r)
    wflow.config.set("input.vertical.g_tt", g_ttm)
    wflow.config.set("input.vertical.kv₀", kv)

    # Convert to v1
    wflow.upgrade_to_v1_wflow()

    # Check with a test config
    config_fn_v1 = join(TESTDATADIR, "wflow_v0x", "sbm", "wflow_sbm_v1.toml")
    wflow_v1 = WflowSbmModel(root, config_filename=config_fn_v1, mode="r")

    # Set kinematic_wave__adaptive_time_step_flag to false to mirror settings in wflow
    #  v1 config
    wflow.config.data["model"]["kinematic_wave__adaptive_time_step_flag"] = False
    assert wflow.config.test_equal(wflow_v1.config)[0]


def test_convert_to_wflow_v1_sediment():
    # Initialize wflow model
    root = join(EXAMPLEDIR, "wflow_upgrade", "sediment")
    config_fn = "wflow_sediment_v0x.toml"

    wflow = WflowSedimentModel(
        root, config_filename=config_fn, data_libs=["artifact_data"], mode="r"
    )
    # Convert to v1
    wflow.upgrade_to_v1_wflow(
        soil_fn="soilgrids",
    )

    # Check with a test config
    config_fn_v1 = join(TESTDATADIR, "wflow_v0x", "sediment", "wflow_sediment_v1.toml")
    wflow_v1 = WflowSedimentModel(root, config_filename=config_fn_v1, mode="r")

    assert wflow.config.test_equal(wflow_v1.config)[0]

    # Checks on extra data in staticmaps
    assert "soil_sagg_fraction" in wflow.staticmaps.data
    assert "land_govers_c" in wflow.staticmaps.data
    assert "river_kodatie_a" in wflow.staticmaps.data
    assert "reservoir_outlet_id" in wflow.staticmaps.data
    res_ids = np.unique(
        wflow.staticmaps.data["reservoir_outlet_id"].raster.mask_nodata()
    )
    assert np.all(np.isin([3349.0, 3367.0, 169986.0], res_ids))
    assert np.all(
        np.isin(
            [0.0, 1.0],
            wflow.staticmaps.data["reservoir_trapping_efficiency"].raster.mask_nodata(),
        )
    )


def test_config_toml_overwrite(tmp_path: Path):
    dummy_model = WflowSbmModel(root=tmp_path, mode="w")
    dummy_model.config.read()
    dummy_model.config.set(
        "input.forcing.khorfrac.value",
        100,
    )
    dummy_model.config.set(
        "input.forcing.khorfrac.value",
        200,
    )
    assert dummy_model.config.get_value("input.forcing.khorfrac.value") == 200

    # Test overwriting top-level key
    dummy_model.config.set("path_log", "log_file.log")
    dummy_model.config.set("path_log", "log_file2.log")
    assert dummy_model.config.get_value("path_log") == "log_file2.log"


def test_convert_to_wflow_v1_with_lake_files(tmp_path: Path):
    # Initialize wflow model
    root = TESTDATADIR / "wflow_v0x" / "sbm_with_lake_files"
    config_fn = "wflow_sbm_v0x.toml"

    wflow = WflowSbmModel(root, config_filename=config_fn, mode="r")

    # Convert to v1
    wflow.upgrade_to_v1_wflow()
    wflow.root.set(tmp_path, mode="w")
    wflow.write()

    assert (tmp_path / "staticmaps" / "reservoir_hq_1.csv").is_file()
    assert (tmp_path / "staticmaps" / "reservoir_hq_2.csv").is_file()



"""Unit tests for hydromt_wflow sediment methods and workflows."""

from os.path import abspath, dirname, join

import geopandas as gpd
import numpy as np

from hydromt_wflow.utils import planar_operation_in_utm

TESTDATADIR = join(dirname(abspath(__file__)), "data")
EXAMPLEDIR = join(dirname(abspath(__file__)), "..", "examples")


def test_setup_lulc_sed(
    example_sediment_model, planted_forest_testdata: gpd.GeoDataFrame
):
    example_sediment_model.setup_lulcmaps(
        lulc_fn="globcover_2009",
        lulc_mapping_fn="globcover_mapping_default",
        planted_forest_fn=planted_forest_testdata,
        lulc_vars=["erosion_usle_c"],
        planted_forest_c=0.0881,
        orchard_name="Orchard",
        orchard_c=0.2188,
    )

    centroid = planar_operation_in_utm(
        planted_forest_testdata, lambda geom: geom.centroid
    )
    da = example_sediment_model.staticmaps.data["erosion_usle_c"].raster.sample(
        centroid
    )

    # Strict equality checking is okay here because no processing is actually happening
    # and we want to make sure we don't add any rounding errors
    assert np.all(da.values == np.array([0.0881, 0.2188]))


def test_setup_lulc_vector(
    example_sediment_model,
    globcover_gdf,
    planted_forest_testdata,
):
    # Test for sediment model
    example_sediment_model.setup_lulcmaps_from_vector(
        lulc_fn=globcover_gdf,
        lulc_mapping_fn="globcover_mapping_default",
        planted_forest_fn=planted_forest_testdata,
        lulc_res=None,
        save_raster_lulc=False,
        planted_forest_c=0.0881,
    )
    assert "erosion_usle_c" in example_sediment_model.staticmaps.data


def test_setup_soilmaps_sed(
    example_sediment_model,
):
    values = example_sediment_model.staticmaps.data[
        "erosion_usle_k"
    ].raster.mask_nodata()
    mean_val = values.mean().values
    assert np.isclose(mean_val, 0.022215, atol=1e-6)

    example_sediment_model.setup_soilmaps(
        soil_fn="soilgrids",
        usle_k_method="epic",
        add_aggregates=False,
    )
    da = example_sediment_model.staticmaps.data

    values = da["erosion_usle_k"].raster.mask_nodata()
    mean_val = values.mean().values
    assert np.isclose(mean_val, 0.0307964, atol=1e-6)

    assert "soil_sediment_d50" in da
    assert "soil_clay_fraction" in da

    soil_composition = (
        da["soil_clay_fraction"].raster.mask_nodata()
        + da["soil_silt_fraction"].raster.mask_nodata()
        + da["soil_sand_fraction"].raster.mask_nodata()
    )
    mask = ~np.isnan(soil_composition.values)
    assert np.allclose(soil_composition.values[mask], 1, atol=1e-6)



"""Unit tests for hydromt_wflow methods and workflows."""

from itertools import product
from os.path import abspath, dirname, join
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr
import xarray.testing as xrt
from hydromt.data_catalog.sources import create_source
from hydromt.gis import GeoDataset, full_like

from hydromt_wflow import workflows
from hydromt_wflow.wflow_sbm import WflowSbmModel

TESTDATADIR = join(dirname(abspath(__file__)), "data")
EXAMPLEDIR = join(dirname(abspath(__file__)), "..", "examples")


def test_setup_basemaps(tmpdir: Path):
    # Region
    region = {
        "basin": [12.2051, 45.8331],
        "strord": 4,
        "bounds": [11.70, 45.35, 12.95, 46.70],
    }
    mod = WflowSbmModel(
        root=tmpdir.join("wflow_base"),
        mode="w",
        data_libs=["artifact_data"],
    )

    hydrography = mod.data_catalog.get_rasterdataset("merit_hydro")
    # Change dtype to uint32
    hydrography["basins"] = hydrography["basins"].astype("uint32")

    # Run setup_basemaps
    mod.setup_basemaps(
        region=region,
        hydrography_fn=hydrography.copy(),
        res=hydrography.raster.res[0],  # no upscaling
    )

    assert mod.staticmaps.data["subcatchment"].dtype == "int32"

    # Test for too small basins
    region = {"subbasin": [12.572061, 46.601984]}
    with pytest.raises(ValueError) as error:  # noqa PT011
        mod.setup_basemaps(
            region=region,
            hydrography_fn=hydrography,
        )
    assert str(error.value).startswith(
        "(Sub)basin at original resolution should at least consist of two cells"
    )

    region = {"subbasin": [12.572061, 46.600359]}
    with pytest.raises(ValueError) as error:  # noqa PT011
        mod.setup_basemaps(
            region=region,
            hydrography_fn=hydrography,
        )
    assert str(error.value).startswith(
        "The output extent at model resolution should at least consist of two cells on"
    )


def test_setup_grid(example_wflow_model):
    # Tests on setup_grid_from_raster
    example_wflow_model.setup_grid_from_raster(
        raster_fn="merit_hydro",
        reproject_method="average",
        variables=["elevtn"],
        wflow_variables=["land_surface_water_flow__ground_elevation"],
        fill_method="nearest",
    )
    assert "elevtn" in example_wflow_model.staticmaps.data
    assert (
        example_wflow_model.config.get_value(
            "input.static.land_surface_water_flow__ground_elevation"
        )
        == "elevtn"
    )

    example_wflow_model.setup_grid_from_raster(
        raster_fn="globcover_2009",
        reproject_method="mode",
    )
    assert "globcover" in example_wflow_model.staticmaps.data

    # Test on exceptions
    with pytest.raises(ValueError, match="Length of variables"):
        example_wflow_model.setup_grid_from_raster(
            raster_fn="merit_hydro",
            reproject_method="average",
            variables=["elevtn", "lndslp"],
            wflow_variables=["land_surface_water_flow__ground_elevation"],
        )
    with pytest.raises(ValueError, match="variables list is not provided"):
        example_wflow_model.setup_grid_from_raster(
            raster_fn="merit_hydro",
            reproject_method="average",
            wflow_variables=["land_surface_water_flow__ground_elevation"],
        )


def test_projected_crs(tmpdir: Path):
    # Instantiate wflow model
    root = str(tmpdir.join("wflow_projected"))
    mod = WflowSbmModel(
        root=root,
        mode="w",
        data_libs=["artifact_data", join(TESTDATADIR, "merit_utm", "merit_utm.yml")],
    )

    # Setup basemaps
    with pytest.raises(
        ValueError,
        match=r"The model resolution 1000 should be smaller than 1 degree",
    ) as error:
        mod.setup_basemaps(
            region={"basin": [12.862, 45.701]},
            res=1000,
            hydrography_fn="merit_hydro",
        )
    assert str(error.value).startswith(
        "The model resolution 1000 should be smaller than 1 degree"
    )

    with pytest.raises(
        ValueError,
        match=r"Model resolution 0.01 should be larger than",
    ) as error:
        mod.setup_basemaps(
            region={"basin": [1427596.0, 5735404.0]},
            res=0.01,
            hydrography_fn="merit_hydro_1k_utm",
            basin_index_fn=None,
        )
    assert str(error.value).startswith("Model resolution 0.01 should be larger than")

    mod.setup_basemaps(
        region={"basin": [1427596.0, 5735404.0]},
        res=2000,
        hydrography_fn="merit_hydro_1k_utm",
        basin_index_fn=None,
    )

    # Add more data eg landuse
    mod.setup_lulcmaps("globcover_2009", lulc_mapping_fn="globcover_mapping_default")

    assert mod.staticmaps.data.raster.crs == 3857
    # 95 quantile is class 190 ie urban
    assert (mod.staticmaps.data["meta_landuse"] == 190).count().values == 338
    assert mod.config.get_value("model.cell_length_in_meter__flag") == True


@pytest.mark.parametrize("glacier_fn", ["glaciers_4326", "glaciers_3857"])
def test_projected_crs_glaciers(glacier_fn, tmpdir):
    # Instantiate wflow model
    root = str(tmpdir.join("wflow_projected"))
    mod = WflowSbmModel(
        root=root,
        mode="w",
        data_libs=[
            "artifact_data",
            join(TESTDATADIR, "merit_utm", "merit_utm.yml"),
            join(TESTDATADIR, "glacier", "glacier_utm.yml"),
        ],
    )

    mod.setup_basemaps(
        region={"basin": [1427596.0, 5735404.0]},
        res=2000,
        hydrography_fn="merit_hydro_1k_utm",
        basin_index_fn=None,
    )

    # Add glaciers
    mod.setup_glaciers(glacier_fn)

    # Confirm glacier maps exist
    assert "meta_glacier_area_id" in mod.staticmaps.data
    assert "glacier_fraction" in mod.staticmaps.data
    assert "glacier_initial_leq_depth" in mod.staticmaps.data
    assert "glaciers" in mod.geoms.data

    # Confirm glaciers have the same CRS as the grid (merit_utm is 3857)
    assert mod.staticmaps.data["glacier_fraction"].raster.crs == 3857
    assert mod.geoms.get("glaciers").crs == 3857

    # Confirm glacier fraction has values
    assert (mod.staticmaps.data["glacier_fraction"] > 0).any().item()

    # Confirm glacier IDs
    assert mod.staticmaps.data["meta_glacier_area_id"].max().item() == 1

    # Confirm config flags
    assert mod.config.get_value("model.glacier__flag") is True
    assert (
        mod.config.get_value("state.variables.glacier_ice__leq_depth")
        == "glacier_leq_depth"
    )


@pytest.fixture
def model_with_rating_curve_data(
    tmpdir: Path, example_wflow_model: WflowSbmModel
) -> tuple[WflowSbmModel, int]:
    # Create dummy lake rating curves
    lakes = example_wflow_model.geoms.get("meta_reservoirs_no_control")
    lake_id = lakes["waterbody_id"].iloc[0]
    area = lakes["reservoir_area"].iloc[0]
    dis = lakes["meta_reservoir_mean_outflow"].iloc[0]
    lvl = lakes["reservoir_initial_depth"].iloc[0]
    elev = lakes["Elevation"].iloc[0]
    lvls = np.linspace(0, lvl)

    df = pd.DataFrame(data={"elevtn": (lvls + elev), "volume": (lvls * area)})
    df = df.join(
        pd.DataFrame(
            {"elevtn": (lvls[-5:-1] + elev), "discharge": np.linspace(0, dis, num=4)}
        ).set_index("elevtn"),
        on="elevtn",
    )
    fn_lake = join(tmpdir, f"rating_curve_{lake_id}.csv")
    df.to_csv(fn_lake, sep=",", index=False, header=True)

    source = create_source(
        data={
            "name": f"lake_rating_test_{lake_id}",
            "data_type": "DataFrame",
            "driver": {
                "name": "pandas",
            },
            "uri": str(tmpdir / f"rating_curve_{lake_id}.csv"),
        }
    )

    # Register as new data source
    example_wflow_model.data_catalog.add_source(
        name=f"lake_rating_test_{lake_id}", source=source
    )
    return example_wflow_model, lake_id


def test_setup_reservoirs_no_control(
    tmpdir: Path, model_with_rating_curve_data: tuple[WflowSbmModel, int]
):
    example_wflow_model, lake_id = model_with_rating_curve_data

    # Update model with it
    example_wflow_model.setup_reservoirs_no_control(
        reservoirs_fn="hydro_lakes",
        rating_curve_fns=[f"lake_rating_test_{lake_id}"],
        min_area=5,
    )

    assert f"reservoir_sh_{lake_id}" in example_wflow_model.tables.data
    assert f"reservoir_hq_{lake_id}" in example_wflow_model.tables.data
    assert 2 in np.unique(
        example_wflow_model.staticmaps.data["reservoir_storage_curve"].values
    )
    assert 1 in np.unique(
        example_wflow_model.staticmaps.data["reservoir_rating_curve"].values
    )
    assert (
        "meta_reservoir_max_storage" not in example_wflow_model.staticmaps.data
    )  # no Vol_max column in hydro_lakes
    for reservoir_layer in [
        "reservoir_rating_curve",
        "reservoir_storage_curve",
        "reservoir_lower_id",
    ]:
        assert example_wflow_model.staticmaps.data[reservoir_layer].dtype == np.int32
        assert isinstance(
            example_wflow_model.staticmaps.data[reservoir_layer]._FillValue, np.int32
        )
    # Write and read back
    new_root = tmpdir / "wflow_lake_test"
    example_wflow_model.root.set(new_root, mode="w")
    example_wflow_model.tables.write()

    test_table = example_wflow_model.tables.data[f"reservoir_sh_{lake_id}"]

    example_wflow_model.tables.data.clear()
    example_wflow_model.root.set(new_root, mode="r")
    example_wflow_model.tables.read()

    assert example_wflow_model.tables.data[f"reservoir_sh_{lake_id}"].equals(test_table)


@pytest.mark.timeout(120)  # max 2 min
@pytest.mark.parametrize("source", ["gww", "jrc"])
def test_reservoirs_simple_control(source, tmpdir, example_wflow_model):
    # Read model 'wflow_piave_subbasin' from EXAMPLEDIR
    model = "wflow"
    example_wflow_model.read()

    # Update model (reservoirs only)
    destination = str(tmpdir.join(model))
    example_wflow_model.root.set(destination, mode="w")

    config = {
        "setup_reservoirs_simple_control": {
            "reservoirs_fn": "hydro_reservoirs",
            "timeseries_fn": source,
            "min_area": 0.0,
            "overwrite_existing": False,
        }
    }

    example_wflow_model.update(model_out=destination, steps=[config], write=False)
    example_wflow_model.write()

    # Check if all parameter maps are available
    required = [
        "reservoir_demand",
        "reservoir_max_release",
        "reservoir_max_volume",
        "reservoir_area",
        "reservoir_target_full_fraction",
        "reservoir_target_min_fraction",
        "reservoir_rating_curve",
        "reservoir_storage_curve",
        "reservoir_initial_depth",
    ]
    assert all(
        x == True
        for x in [k in example_wflow_model.staticmaps.data.keys() for k in required]
    ), "1 or more reservoir map missing"

    # Check if all parameter maps contain x non-null values, where x equals
    # the number of reservoirs in the model area
    grid = example_wflow_model.staticmaps.data.where(
        example_wflow_model.staticmaps.data.reservoir_outlet_id != -999
    )
    stacked = grid.reservoir_outlet_id.stack(x=[grid.raster.y_dim, grid.raster.x_dim])
    stacked = stacked[stacked.notnull()]
    number_of_reservoirs = stacked.size

    for i in required:
        assert (
            np.count_nonzero(
                ~np.isnan(
                    grid[i].sel(
                        {
                            grid.raster.y_dim: stacked[grid.raster.y_dim].values,
                            grid.raster.x_dim: stacked[grid.raster.x_dim].values,
                        }
                    )
                )
            )
            == number_of_reservoirs
        ), (
            f"Number of non-null values in {i} not equal to \
number of reservoirs in model area"
        )


def test_setup_ksathorfrac(tmpdir, example_wflow_model):
    # Read the modeldata
    model = "wflow"
    example_wflow_model.read()
    # Create dummy ksat data
    da = full_like(example_wflow_model.staticmaps.data["soil_ksat_vertical"])
    data = np.zeros(da.shape)
    for x, y in product(*[range(item) for item in da.shape]):
        data[x, y] = 750 - ((x + y) ** 0.4 * 114.07373)
    da.values = data
    da.name = "subsurface_ksat_horizontal_ratio"
    # Set the output directory
    destination = str(tmpdir.join(model))
    example_wflow_model.root.set(destination, mode="w")

    # Build the map
    example_wflow_model.setup_ksathorfrac(
        ksat_fn=da,
    )

    # Check values
    values = example_wflow_model.staticmaps.data[
        "subsurface_ksat_horizontal_ratio"
    ].raster.mask_nodata()
    max_val = values.max().values
    mean_val = values.mean().values
    assert np.isclose(max_val, 431.75)
    assert np.isclose(mean_val, 220.206)


def test_setup_ksatver_vegetation(example_wflow_model):
    # Build the soil_ksat_vertical vegetation map
    example_wflow_model.setup_ksatver_vegetation(
        soil_fn="soilgrids",
        output_name="soil_ksat_vertical_vegetation_test",
    )

    # Check values
    values = example_wflow_model.staticmaps.data[
        "soil_ksat_vertical_vegetation_test"
    ].raster.mask_nodata()
    max_val = values.max().values
    mean_val = values.mean().values
    assert int(max_val) == 4247
    assert int(mean_val) == 1672


def test_setup_lai(example_wflow_model: WflowSbmModel):
    # Use vito and MODIS lai data for testing
    # Read vegetation_leaf_area_index data
    da_lai = example_wflow_model.data_catalog.get_rasterdataset(
        "modis_lai", geom=example_wflow_model.region, buffer=2
    )
    # Read landuse data
    da_landuse = example_wflow_model.data_catalog.get_rasterdataset(
        "vito_2015", geom=example_wflow_model.region, buffer=2
    )

    # Derive mapping using the method any
    df_lai_any = workflows.create_lulc_lai_mapping_table(
        da_lulc=da_landuse.copy(),
        da_lai=da_lai.copy(),
        sampling_method="any",
        lulc_zero_classes=[80, 200, 0],
    )
    # Check that all landuse classes are present in the mapping
    assert np.all(df_lai_any.index.values == np.unique(da_landuse))

    # Try with the other two methods
    df_lai_mode = workflows.create_lulc_lai_mapping_table(
        da_lulc=da_landuse.copy(),
        da_lai=da_lai.copy(),
        sampling_method="mode",
        lulc_zero_classes=[80, 200, 0],
    )
    df_lai_q3 = workflows.create_lulc_lai_mapping_table(
        da_lulc=da_landuse.copy(),
        da_lai=da_lai.copy(),
        sampling_method="q3",
        lulc_zero_classes=[80, 200, 0],
    )
    # Check the number of landuse classes in the mapping tables
    assert len(df_lai_any[df_lai_any.samples == 0]) == 1
    assert len(df_lai_mode[df_lai_mode.samples == 0]) == 3
    assert len(df_lai_q3[df_lai_q3.samples == 0]) == 3
    # Check number of samples for landuse class 20 with the different methods
    assert int(df_lai_any.loc[20].samples) == 2481
    assert int(df_lai_mode.loc[20].samples) == 59
    assert int(df_lai_q3.loc[20].samples) == 4

    # Try to use the mapping tables to setup the vegetation_leaf_area_index
    example_wflow_model.setup_laimaps_from_lulc_mapping(
        lulc_fn="vito_2015",
        lai_mapping_fn=df_lai_any,
        output_name="lai_from_vito_mapping",
    )

    assert "lai_from_vito_mapping" in example_wflow_model.staticmaps.data
    assert (
        example_wflow_model.config.get_value("input.cyclic.vegetation__leaf_area_index")
        == "lai_from_vito_mapping"
    )


def test_setup_rootzoneclim(example_wflow_model):
    # load csv with dummy data for long timeseries of precip, pet and dummy Q data.
    test_data = pd.read_csv(
        join(TESTDATADIR, "df_sub_dummy.csv"),
        parse_dates=True,
        dayfirst=True,
        index_col=0,
        header=0,
    )

    # open too short time series of eobs from artifact data - to get example extent.
    ds_obs = example_wflow_model.data_catalog.get_rasterdataset(
        "eobs",
        geom=example_wflow_model.region,
        buffer=2,
        variables=["temp", "precip"],
    )

    # use long dummy timeseries of pet, precip and Q in a dataset
    zeros = np.zeros((len(test_data), len(ds_obs.latitude), len(ds_obs.longitude)))
    ds = xr.Dataset(
        data_vars=dict(
            precip=(["time", "latitude", "longitude"], zeros),
            pet=(["time", "latitude", "longitude"], zeros),
        ),
        coords=dict(
            time=test_data.index,
            latitude=ds_obs.latitude.values,
            longitude=ds_obs.longitude.values,
        ),
    )
    # fill data with dummy precip and pet data - uniform over area
    ds["pet"] = ds["pet"] + test_data["pet"].to_xarray()
    ds["precip"] = ds["precip"] + test_data["precip"].to_xarray()
    ds.raster.set_crs(ds_obs.raster.crs)

    ds_cc_hist = ds.copy()
    ds_cc_hist["pet"] = ds["pet"] * 1.05

    ds_cc_fut = ds.copy()
    ds_cc_fut["pet"] = ds["pet"] * 1.3

    # make dummy dataset with dummy Q data
    indices = [1, 2, 3]
    ds_run = xr.Dataset(
        data_vars=dict(
            discharge=(
                ["time", "index"],
                np.zeros((len(test_data.index), len(indices))),
            ),
        ),
        coords=dict(
            time=test_data.index,
            index=indices,
            lon=(
                ["index"],
                example_wflow_model.geoms.get("gauges_grdc").geometry.x.values,
            ),
            lat=(
                ["index"],
                example_wflow_model.geoms.get("gauges_grdc").geometry.y.values,
            ),
        ),
    )
    # ds_run.raster.set_crs(ds_obs.raster.crs)
    # ds_run.raster.set_spatial_dims(x_dim="lon", y_dim="lat")
    # fill with dummy data
    for ind in indices:
        ds_run["discharge"].loc[dict(index=ind)] = test_data[f"Q_{ind}"]

    example_wflow_model.setup_rootzoneclim(
        run_fn=ds_run,
        forcing_obs_fn=ds,
        forcing_cc_hist_fn=ds_cc_hist,
        forcing_cc_fut_fn=ds_cc_fut,
        start_hydro_year="Oct",
        start_field_capacity="Apr",
        time_range=("2005-01-01", "2020-12-31"),
        time_range_fut=("2005-01-01", "2020-12-31"),
        missing_days_threshold=330,
        return_period=[2, 5, 10, 15, 20],
        output_name_rootingdepth="vegetation_root_depth_obs_2",
        rootzone_storage=True,
    )

    assert "vegetation_root_depth_obs_20" in example_wflow_model.staticmaps.data
    assert "vegetation_root_depth_cc_hist_20" in example_wflow_model.staticmaps.data
    assert "vegetation_root_depth_cc_fut_20" in example_wflow_model.staticmaps.data

    assert "meta_rootzone_storage_obs_15" in example_wflow_model.staticmaps.data
    assert "meta_rootzone_storage_cc_hist_15" in example_wflow_model.staticmaps.data
    assert "meta_rootzone_storage_cc_fut_15" in example_wflow_model.staticmaps.data

    assert (
        example_wflow_model.config.get_value("input.static.vegetation_root__depth")
        == "vegetation_root_depth_obs_2"
    )

    assert example_wflow_model.geoms.get("rootzone_storage").loc[1][
        "rootzone_storage_obs_2"
    ] == pytest.approx(72.32875652734612, abs=0.5)
    assert example_wflow_model.geoms.get("rootzone_storage").loc[1][
        "rootzone_storage_cc_hist_2"
    ] == pytest.approx(70.26993058321949, abs=0.5)
    assert example_wflow_model.geoms.get("rootzone_storage").loc[1][
        "rootzone_storage_cc_fut_2"
    ] == pytest.approx(80.89081789601374, abs=0.5)

    # change settings for LAI and correct_cc_deficit
    example_wflow_model.setup_rootzoneclim(
        run_fn=ds_run,
        forcing_obs_fn=ds,
        forcing_cc_hist_fn=ds_cc_hist,
        forcing_cc_fut_fn=ds_cc_fut,
        LAI=True,
        start_hydro_year="Oct",
        start_field_capacity="Apr",
        time_range=("2005-01-01", "2015-12-31"),
        time_range_fut=("2005-01-01", "2015-12-31"),
        correct_cc_deficit=True,
        missing_days_threshold=330,
        return_period=[2, 5, 10, 15, 20],
        output_name_rootingdepth="vegetation_root_depth_obs_2",
    )

    assert example_wflow_model.geoms.get("rootzone_storage").loc[1][
        "rootzone_storage_obs_2"
    ] == pytest.approx(82.85684577620462, abs=0.5)
    assert example_wflow_model.geoms.get("rootzone_storage").loc[1][
        "rootzone_storage_cc_hist_2"
    ] == pytest.approx(82.44039441508069, abs=0.5)
    assert example_wflow_model.geoms.get("rootzone_storage").loc[1][
        "rootzone_storage_cc_fut_2"
    ] == pytest.approx(106.03809681174451, abs=0.5)


def test_setup_outlets(example_wflow_model):
    # Update subcatchment ID
    new_subcatch = example_wflow_model.staticmaps.data["subcatchment"].copy()
    new_subcatch = new_subcatch.where(new_subcatch == new_subcatch.raster.nodata, 1001)
    example_wflow_model.staticmaps.set(new_subcatch, "subcatchment")

    # Derive outlets
    example_wflow_model.setup_outlets()

    # Check if the ID is indeed 1001
    val, count = np.unique(
        example_wflow_model.staticmaps.data["outlets"], return_counts=True
    )
    # 0 is no data
    assert val[1] == 1001
    assert count[1] == 1


def test_setup_gauges(example_wflow_model: WflowSbmModel):
    # 1. Test with grdc data
    # uparea rename not in the latest artifact_data version
    source = example_wflow_model.data_catalog.get_source("grdc")
    source.data_adapter.rename = {"area": "uparea"}

    example_wflow_model.setup_gauges(
        gauges_fn="grdc",
        basename="grdc_uparea",
        snap_to_river=False,
        mask=None,
        snap_uparea=True,
        wdw=5,
        rel_error=0.05,
    )
    gdf = example_wflow_model.geoms.get("gauges_grdc_uparea")
    ds_samp = example_wflow_model.staticmaps.data[
        ["river_mask", "meta_upstream_area"]
    ].raster.sample(gdf, wdw=0)
    assert np.allclose(
        ds_samp["meta_upstream_area"].values,
        gdf["uparea"].values,
        rtol=0.05,
    )

    # 2. Test with/without snapping to mask
    stations_fn = join(TESTDATADIR, "test_stations.csv")
    example_wflow_model.setup_gauges(
        gauges_fn=stations_fn,
        basename="stations_snapping",
        snap_to_river=True,
        mask=None,
    )
    gdf_snap = example_wflow_model.geoms.get("gauges_stations_snapping")

    example_wflow_model.setup_gauges(
        gauges_fn=stations_fn,
        basename="stations_no_snapping",
        snap_to_river=False,
        mask=None,
    )
    gdf_no_snap = example_wflow_model.geoms.get("gauges_stations_no_snapping")

    # Check that not all geometries of gdf_snap and gdf_no_snap are equal
    assert not gdf_snap.equals(gdf_no_snap)
    # Find which row is identical
    equal = gdf_snap[gdf_snap["geometry"] == gdf_no_snap["geometry"]]
    assert len(equal) == 1
    assert equal.index.values[0] == 1003

    # 3. Test uparea with/without river snapping
    example_wflow_model.setup_gauges(
        gauges_fn=stations_fn,
        basename="stations_uparea_no_snapping",
        snap_to_river=False,
        mask=None,
        snap_uparea=True,
        wdw=5,
        rel_error=0.05,
        fillna=False,
    )
    gdf_no_snap = example_wflow_model.geoms.get("gauges_stations_uparea_no_snapping")
    # Only two gauges have uparea values and fillna is False
    assert gdf_no_snap.index.size == 2

    example_wflow_model.setup_gauges(
        gauges_fn=stations_fn,
        basename="stations_uparea_no_snapping_fillna",
        snap_to_river=False,
        mask=None,
        snap_uparea=True,
        wdw=5,
        rel_error=0.05,
        fillna=True,
    )
    gdf_no_snap_fillna = example_wflow_model.geoms.get(
        "gauges_stations_uparea_no_snapping_fillna"
    )
    # Two gauges have uparea values and fillna is True
    assert gdf_no_snap_fillna.index.size == 3
    # Not all gauges are in the river as snap_to_river is False
    ds_samp = example_wflow_model.staticmaps.data[
        ["river_mask", "meta_upstream_area"]
    ].raster.sample(gdf_no_snap_fillna, wdw=0)
    assert not np.all(ds_samp["river_mask"].values == 1)

    example_wflow_model.setup_gauges(
        gauges_fn=stations_fn,
        basename="stations_uparea_snapping",
        snap_to_river=True,
        mask=None,
        snap_uparea=True,
        wdw=5,
        rel_error=0.05,
        abs_error=25,
        fillna=False,
    )
    gdf_snap = example_wflow_model.geoms.get("gauges_stations_uparea_snapping")
    # Only one gauge has uparea value and is in the river
    # (the one with NaN for upstream area would have ended in the river if fillna=True)
    assert gdf_snap.index.size == 1
    # Check that they are all in the river
    ds_samp = example_wflow_model.staticmaps.data[
        ["river_mask", "meta_upstream_area"]
    ].raster.sample(gdf_snap, wdw=0)
    assert np.all(ds_samp["river_mask"].values == 1)

    # 4. Test with test_stations.csv file
    stations_csv_file = join(EXAMPLEDIR, "test_stations.csv")
    example_wflow_model.setup_gauges(gauges_fn=stations_csv_file, basename="test-flow")
    test_stations_gdf = example_wflow_model.geoms.get("gauges_test-flow")
    assert len(test_stations_gdf) == 3


@pytest.mark.parametrize("elevtn_map", ["land_elevation", "meta_subgrid_elevation"])
def test_setup_rivers(elevtn_map, floodplain1d_testdata, example_wflow_model):
    example_wflow_model.setup_rivers(
        hydrography_fn="merit_hydro",
        river_geom_fn="hydro_rivers_lin",
        river_upa=30,
        rivdph_method="powlaw",
        min_rivdph=1,
        min_rivwth=30,
        slope_len=2000,
        smooth_len=5000,
        river_routing="local_inertial",
        elevtn_map=elevtn_map,
        output_names={},
    )

    mapname = {
        "land_elevation": "river_bank_elevation_avg",
        "meta_subgrid_elevation": "river_bank_elevation_subgrid",
    }[elevtn_map]

    assert mapname in example_wflow_model.staticmaps.data
    assert (
        example_wflow_model.config.get_value("model.river_routing") == "local_inertial"
    )
    assert (
        example_wflow_model.config.get_value("input.static.river_bank_water__elevation")
        == mapname
    )
    assert (
        example_wflow_model.staticmaps.data[mapname]
        .raster.mask_nodata()
        .equals(floodplain1d_testdata[mapname])
    )


def test_setup_rivers_no_subgrid(tmpdir: Path):
    # Instantiate new wflow model
    # Region
    region = {
        "subbasin": [12.5032, 46.5327],
        "strord": 4,
        "bounds": [11.70, 45.35, 12.95, 46.70],
    }
    mod = WflowSbmModel(
        root=str(tmpdir.join("river_no_subgrid")),
        mode="w",
        data_libs=["artifact_data"],
    )
    hydrography = mod.data_catalog.get_rasterdataset("merit_hydro_ihu")
    # Run setup_basemaps
    mod.setup_basemaps(
        region=region,
        hydrography_fn=hydrography.copy(),
        res=hydrography.raster.res[0],  # no upscaling
    )

    assert "x_out" not in mod.staticmaps.data

    # Setup rivers with different hydrography (with subgrid = false)
    with pytest.raises(ValueError, match="It seems model grid was not upscaled"):
        mod.setup_rivers(
            hydrography_fn="merit_hydro",
            river_geom_fn="hydro_rivers_lin",
            river_upa=30,
        )

    # Now with correct hydrography
    mod.setup_rivers(
        hydrography_fn="merit_hydro_ihu",
        river_geom_fn="hydro_rivers_lin",
        river_upa=30,
    )

    assert "river_mask" in mod.staticmaps.data


def test_setup_rivers_depth(tmpdir: Path):
    # Instantiate new wflow model
    # Region
    region = {
        "subbasin": [12.3661, 46.3998],
        "strord": 4,
        "bounds": [11.70, 45.35, 12.95, 46.70],
    }
    mod = WflowSbmModel(
        root=str(tmpdir.join("river_mask")),
        mode="w",
        data_libs=["artifact_data"],
    )

    mod.setup_basemaps(
        region=region,
        hydrography_fn="merit_hydro",
        basin_index_fn="merit_hydro_index",
        res=0.016666,
    )

    # Try using manning method
    mod.setup_rivers(
        hydrography_fn="merit_hydro",
        river_geom_fn="hydro_rivers_lin",
        river_upa=30,
        rivdph_method="manning",
        min_rivdph=1,
        min_rivwth=30,
        slope_len=2000,
        smooth_len=5000,
        river_routing="local_inertial",
        elevtn_map="land_elevation",
    )

    assert "river_depth" in mod.staticmaps.data

    # Try using gvf method
    mod.setup_rivers(
        hydrography_fn="merit_hydro",
        river_geom_fn="hydro_rivers_lin",
        river_upa=30,
        rivdph_method="gvf",
        min_rivdph=1,
        min_rivwth=30,
        slope_len=2000,
        smooth_len=5000,
        river_routing="local_inertial",
        elevtn_map="meta_subgrid_elevation",
    )

    # RiverDepth iteslf doesn't matter here, this assertion
    # is just to check the method ran without errors
    # as this will error if something went wrong in the process
    assert "river_depth" in mod.staticmaps.data


def test_setup_floodplains_1d(
    example_wflow_model: WflowSbmModel, floodplain1d_testdata
):
    flood_depths = [0.5, 1.0, 1.5, 2.0, 2.5]

    example_wflow_model.setup_rivers(
        hydrography_fn="merit_hydro",
        river_geom_fn="hydro_rivers_lin",
        river_upa=30,
        rivdph_method="powlaw",
        min_rivdph=1,
        min_rivwth=30,
        slope_len=2000,
        smooth_len=5000,
        river_routing="local_inertial",
        elevtn_map="land_elevation",
    )

    example_wflow_model.setup_floodplains(
        hydrography_fn="merit_hydro",
        floodplain_type="1d",
        river_upa=30,
        flood_depths=flood_depths,
    )

    assert "floodplain_volume" in example_wflow_model.staticmaps.data
    assert example_wflow_model.config.get_value("model.floodplain_1d__flag") == True
    assert (
        example_wflow_model.config.get_value("model.land_routing") == "kinematic_wave"
    )
    assert (
        example_wflow_model.config.get_value(
            "input.static.floodplain_water__sum_of_volume_per_depth"
        )
        == "floodplain_volume"
    )
    name = example_wflow_model._MAPS["floodplain_volume"]
    assert np.all(
        example_wflow_model.staticmaps.data[name].flood_depth.values == flood_depths
    )
    xrt.assert_allclose(
        example_wflow_model.staticmaps.data[name].raster.mask_nodata(),
        floodplain1d_testdata.floodplain_volume,
    )

    # Check states in config
    assert (
        example_wflow_model.config.get_value(
            "state.variables.floodplain_water__instantaneous_volume_flow_rate"
        )
        == "floodplain_instantaneous_q"
    )

    assert (
        example_wflow_model.config.get_value("state.variables.floodplain_water__depth")
        == "floodplain_h"
    )

    assert (
        example_wflow_model.config.get_value(
            "state.variables.land_surface_water__instantaneous_volume_flow_rate"
        )
        == "land_instantaneous_q"
    )


@pytest.mark.parametrize("elevtn_map", ["land_elevation", "meta_subgrid_elevation"])
def test_setup_floodplains_2d(
    elevtn_map, example_wflow_model: WflowSbmModel, floodplain1d_testdata
):
    example_wflow_model.setup_rivers(
        hydrography_fn="merit_hydro",
        river_geom_fn="hydro_rivers_lin",
        river_upa=30,
        rivdph_method="powlaw",
        min_rivdph=1,
        min_rivwth=30,
        slope_len=2000,
        smooth_len=5000,
        river_routing="local_inertial",
        elevtn_map="land_elevation",
        output_names={},
    )

    example_wflow_model.setup_floodplains(
        hydrography_fn="merit_hydro", floodplain_type="2d", elevtn_map=elevtn_map
    )

    mapname = {
        "land_elevation": "river_bank_elevation_avg",
        "meta_subgrid_elevation": "river_bank_elevation_subgrid",
    }[elevtn_map]

    assert f"{mapname}_D4" in example_wflow_model.staticmaps.data
    assert example_wflow_model.config.get_value("model.floodplain_1d__flag") == False
    assert (
        example_wflow_model.config.get_value("model.land_routing") == "local_inertial"
    )
    assert (
        example_wflow_model.config.get_value("input.static.river_bank_water__elevation")
        == f"{mapname}_D4"
    )
    assert (
        example_wflow_model.config.get_value(
            "input.static.land_surface_water_flow__ground_elevation"
        )
        == elevtn_map
    )
    xrt.assert_allclose(
        example_wflow_model.staticmaps.data[f"{mapname}_D4"].raster.mask_nodata(),
        floodplain1d_testdata[f"{mapname}_D4"],
    )

    # Check states in config
    assert (
        example_wflow_model.config.get_value(
            "state.variables.land_surface_water__x_component_of_instantaneous_volume_flow_rate"
        )
        == "land_instantaneous_qx"
    )
    assert (
        example_wflow_model.config.get_value(
            "state.variables.land_surface_water__y_component_of_instantaneous_volume_flow_rate"
        )
        == "land_instantaneous_qy"
    )


def test_setup_precip_from_point_timeseries(
    example_wflow_model: WflowSbmModel, df_precip_stations, gdf_precip_stations
):
    # Interpolation types and the mean value to check the test
    # first value is for all 8 stations, second for 3 stations inside Piave
    # Linear interpolation can result in missing values when coverage is too low
    # such as in the test case (resulting in low value of 155).
    # Note: start_time and end_time of model are used to slice timeseries

    interp_types = {
        "nearest": [446, 435],
        "linear": [459, 155],  # note the low value for linear interpolation
        "idw": [443, 431],
        "ordinarykriging": [446, 426],
        # "externaldriftkriging" requires an additional drift variable,
    }

    for interp_type, test_val in interp_types.items():
        example_wflow_model.setup_precip_from_point_timeseries(
            precip_fn=df_precip_stations,
            precip_stations_fn=gdf_precip_stations,
            interp_type=interp_type,
            buffer=1e6,
        )
        # Check forcing and dtype
        assert "precip" in example_wflow_model.forcing.data
        assert example_wflow_model.forcing.data["precip"].dtype == "float32"

        # Compare computed value with expected value using all stations
        mean_all = example_wflow_model.forcing.data["precip"].mean().values
        assert int(mean_all * 1000) == test_val[0]

        # Do the some but only for the 3 stations inside the basin (buffer = 0)
        example_wflow_model.setup_precip_from_point_timeseries(
            precip_fn=df_precip_stations,
            precip_stations_fn=gdf_precip_stations,
            interp_type=interp_type,
            buffer=0,
        )
        mean_inside = example_wflow_model.forcing.data["precip"].mean().values
        assert int(mean_inside * 1000) == test_val[1]

    # Similar test but for GeoDataset
    for interp_type, test_val in interp_types.items():
        geodataset_precip_stations = GeoDataset.from_gdf(
            gdf=gdf_precip_stations,
            data_vars={"precip": df_precip_stations},
            index_dim="index",
            keep_cols=False,
            merge_index="gdf",
        )

        example_wflow_model.setup_precip_from_point_timeseries(
            precip_fn=geodataset_precip_stations,
            precip_stations_fn=None,
            interp_type=interp_type,
            buffer=1e6,
        )
        # Check forcing and dtype
        assert "precip" in example_wflow_model.forcing.data
        assert example_wflow_model.forcing.data["precip"].dtype == "float32"

        # Compare computed value with expected value using all stations
        mean_all = example_wflow_model.forcing.data["precip"].mean().values
        assert int(mean_all * 1000) == test_val[0]

    # Also include a test for uniform precipitation
    example_wflow_model.setup_precip_from_point_timeseries(
        precip_fn=df_precip_stations.iloc[:, 0].to_frame(),
        precip_stations_fn=None,
        interp_type="uniform",
    )
    # Check if the values per timestep are unique
    for i, _ in enumerate(example_wflow_model.forcing.data["precip"].time):
        unique_values = np.unique(
            example_wflow_model.forcing.data["precip"].isel(time=i)
        )
        assert len(unique_values[~np.isnan(unique_values)]) == 1
    # Check mean value
    mean_uniform = example_wflow_model.forcing.data["precip"].mean().values
    assert int(mean_uniform * 1000) == 274


def test_setup_pet_forcing(example_wflow_model: WflowSbmModel, da_pet: xr.DataArray):
    example_wflow_model.setup_pet_forcing(
        pet_fn=da_pet,
    )

    assert "pet" in example_wflow_model.forcing.data
    # Check dtype
    assert example_wflow_model.forcing.data["pet"].dtype == "float32"
    # used to be debruin before update
    assert "pet_method" not in example_wflow_model.forcing.data["pet"].attrs
    assert example_wflow_model.forcing.data["pet"].min().values == da_pet.min().values
    mean_val = example_wflow_model.forcing.data["pet"].mean().values
    assert int(mean_val * 1000) == 2984


def test_setup_1dmodel_connection(example_wflow_model: WflowSbmModel, rivers1d):
    # test subbasin_area method with river boundaries
    example_wflow_model.setup_1dmodel_connection(
        river1d_fn=rivers1d,
        connection_method="subbasin_area",
        area_max=10.0,
        add_tributaries=True,
        include_river_boundaries=True,
        mapname="1dmodel",
        update_toml=True,
        toml_output="netcdf_scalar",
    )

    assert "gauges_1dmodel" in example_wflow_model.geoms.data
    assert "subcatchment_1dmodel" in example_wflow_model.geoms.data
    assert "subcatchment_riv_1dmodel" in example_wflow_model.geoms.data

    assert len(example_wflow_model.geoms.get("gauges_1dmodel")) == 3
    assert len(example_wflow_model.geoms.get("subcatchment_1dmodel")) == 2
    conf_dict = {
        "name": "Q",
        "map": "gauges_1dmodel",
        "parameter": "river_water__volume_flow_rate",
    }
    assert (
        conf_dict
        in example_wflow_model.config.data["output"]["netcdf_scalar"]["variable"]  # noqa: E501
    )

    # test subbasin_area method with river boundaries
    example_wflow_model.setup_1dmodel_connection(
        river1d_fn=rivers1d,
        connection_method="subbasin_area",
        area_max=30.0,
        add_tributaries=True,
        include_river_boundaries=False,
        mapname="1dmodel-nobounds",
        update_toml=True,
        toml_output="csv",
    )

    assert len(example_wflow_model.geoms.get("gauges_1dmodel-nobounds")) == 1
    assert len(example_wflow_model.geoms.get("subcatchment_1dmodel-nobounds")) == 2
    assert np.all(
        example_wflow_model.geoms.get("subcatchment_1dmodel").geometry.geom_equals(
            example_wflow_model.geoms.get("subcatchment_1dmodel-nobounds").geometry
        )
    )

    # test nodes method without extra tributaries
    example_wflow_model.setup_1dmodel_connection(
        river1d_fn=rivers1d,
        connection_method="nodes",
        add_tributaries=False,
        include_river_boundaries=False,
        mapname="1dmodel-nodes",
        update_toml=False,
        max_dist=5000,
    )

    assert "gauges_1dmodel-nodes" not in example_wflow_model.geoms.data
    assert len(example_wflow_model.geoms.get("subcatchment_1dmodel-nodes")) == 6


def test_setup_1dmodel_connection_other(
    example_wflow_model: WflowSbmModel, rivers1d_projected: gpd.GeoDataFrame
):
    # rivers1d_projected has different crs and parts outside the model area

    # raise ValueError too small area_max
    with pytest.raises(
        ValueError, match=r"^1 subbasin\(s\) do not contain a wflow river cell."
    ):
        example_wflow_model.setup_1dmodel_connection(
            river1d_fn=rivers1d_projected,
            connection_method="subbasin_area",
            area_max=30.0,
            add_tributaries=True,
            include_river_boundaries=True,
            mapname="1dmodel",
            update_toml=True,
            toml_output="netcdf_scalar",
        )

    # Check the coupling and results
    example_wflow_model.setup_1dmodel_connection(
        river1d_fn=rivers1d_projected,
        connection_method="subbasin_area",
        area_max=10.0,
        add_tributaries=True,
        include_river_boundaries=True,
        mapname="1dmodel",
        update_toml=True,
        toml_output="netcdf_scalar",
    )

    assert "gauges_1dmodel" in example_wflow_model.geoms.data
    assert "subcatchment_1dmodel" in example_wflow_model.geoms.data
    assert "subcatchment_riv_1dmodel" in example_wflow_model.geoms.data

    assert len(example_wflow_model.geoms.data["gauges_1dmodel"]) == 1
    assert len(example_wflow_model.geoms.data["subcatchment_1dmodel"]) == 1


def test_skip_nodata_reservoir(clipped_wflow_model: WflowSbmModel):
    # Using the clipped_wflow_model as the reservoirs are not in this model
    clipped_wflow_model.setup_reservoirs_simple_control(
        reservoirs_fn="hydro_reservoirs",
        min_area=0.0,
        overwrite_existing=True,
    )
    assert clipped_wflow_model.config.data["model"]["reservoir__flag"] == False
    # Get names for two reservoir layers
    for mapname in ["reservoir_area_id", "reservoir_outlet_id"]:
        # Check if layers are indeed not present in the model
        assert (
            clipped_wflow_model._MAPS[mapname]
            not in clipped_wflow_model.staticmaps.data.data_vars
        )


def test_setup_lulc_vector(
    example_wflow_model: WflowSbmModel,
    globcover_gdf,
):
    # Test for wflow sbm
    # Use a file directly for lulc_mapping_fn
    mapping_fn = (
        Path(abspath(__file__)).parent.parent
        / "hydromt_wflow"
        / "data"
        / "lulc"
        / "globcover_mapping.csv"
    )
    example_wflow_model.setup_lulcmaps_from_vector(
        lulc_fn=globcover_gdf,
        lulc_mapping_fn=mapping_fn,
        lulc_res=0.0025,
        save_raster_lulc=False,
    )
    assert "meta_landuse" in example_wflow_model.staticmaps.data


def test_setup_lulc_paddy(example_wflow_model: WflowSbmModel, tmpdir: Path):
    # Read the data
    example_wflow_model.read()
    example_wflow_model.root.set(Path(tmpdir), mode="w")

    layers = [50, 100, 50, 200, 800]

    # Use the method
    # Note: 11 this not NOT a rice class, but chosen just for testing purposes
    example_wflow_model.setup_lulcmaps_with_paddy(
        lulc_fn="glcnmo",
        paddy_class=11,
        wflow_thicknesslayers=layers,
    )

    # Set to shorter name to improve readability of tests
    ds = example_wflow_model.staticmaps.data.copy()

    assert "soil_ksat_vertical_factor" in ds
    assert "vegetation_crop_factor" in ds
    assert "soil_brooks_corey_c" in ds
    # Assert layers are updated
    assert example_wflow_model.config.data["model"]["soil_layer__thickness"] == layers
    # Adding +1 to the layers to also represent the last layer
    assert len(ds.layer) == len(layers) + 1
    assert ds.soil_brooks_corey_c.shape[0] == len(layers) + 1
    assert ds.soil_ksat_vertical_factor.shape[0] == len(layers) + 1
    # Assert soil_ksat_vertical_factor is written to vertical section in config
    assert (
        example_wflow_model.config.data["input"]["static"][
            "soil_water__vertical_saturated_hydraulic_conductivity_factor"
        ]
        == "soil_ksat_vertical_factor"
    )

    # Test soil_ksat_vertical_factor is not 1 at the right layer for a paddy cell
    kvfrac_values = ds.soil_ksat_vertical_factor.sel(
        latitude=45.89, longitude=12.10, method="nearest"
    ).values
    assert kvfrac_values[0] == 1.0
    assert kvfrac_values[2] != 1.0
    assert kvfrac_values[5] == 1.0

    # Test values for updated C
    c_values = ds.soil_brooks_corey_c.sel(
        latitude=45.89, longitude=12.10, method="nearest"
    ).values
    assert np.isclose(c_values[0], 9.220022)
    assert np.isclose(c_values[2], 9.553196)
    assert np.isclose(c_values[5], 9.849495)

    # Test values for crop coefficient
    assert np.isclose(
        ds["vegetation_crop_factor"].raster.mask_nodata().mean().values, 0.8869253
    )

    # Test with a separate paddy_map
    example_wflow_model.setup_lulcmaps_with_paddy(
        lulc_fn="globcover_2009",
        paddy_class=11,
        output_paddy_class=12,
        paddy_fn=ds["meta_landuse"].where(
            ds["meta_landuse"] == 11, ds["meta_landuse"].raster.nodata
        ),
        lulc_mapping_fn="globcover_mapping_default",
        wflow_thicknesslayers=layers,
    )

    ds2 = example_wflow_model.staticmaps.data.copy()

    assert np.any(ds2["meta_landuse"] == 12)


def test_setup_allocation_areas(example_wflow_model: WflowSbmModel, tmpdir: Path):
    # Read the data and set new root
    example_wflow_model.read()
    example_wflow_model.root.set(
        Path(
            tmpdir,
        ),
        mode="w",
    )

    # Use the method
    example_wflow_model.setup_allocation_areas(
        waterareas_fn="gadm_level2",
        priority_basins=True,
    )

    # Assert entries
    assert "demand_allocation_area_id" in example_wflow_model.geoms.data
    assert "demand_allocation_area_id" in example_wflow_model.staticmaps.data

    # Assert output values
    assert len(example_wflow_model.geoms.get("demand_allocation_area_id")) == 3
    # on unique values
    uni = example_wflow_model.geoms.get("demand_allocation_area_id").value.unique()
    assert np.all(np.sort(uni) == [11, 16, 17])


def test_setup_allocation_surfacewaterfrac(
    example_wflow_model: WflowSbmModel, tmpdir: Path
):
    # Read the data and set new root
    example_wflow_model.read()
    example_wflow_model.root.set(
        Path(
            tmpdir,
        ),
        mode="w",
    )
    # Add lisflood data from test
    lisflood_yml = join(TESTDATADIR, "demand", "data_catalog.yml")
    example_wflow_model.data_catalog = example_wflow_model.data_catalog.from_yml(
        lisflood_yml
    )

    # Use the method fully with lisflood
    example_wflow_model.setup_allocation_surfacewaterfrac(
        gwfrac_fn="lisflood_gwfrac",
        waterareas_fn="lisflood_waterregions",
        gwbodies_fn="lisflood_gwbodies",
        ncfrac_fn="lisflood_ncfrac",
        interpolate_nodata=False,
    )

    # Assert entries
    assert "demand_surface_water_ratio" in example_wflow_model.staticmaps.data
    assert np.isclose(
        example_wflow_model.staticmaps.data["demand_surface_water_ratio"]
        .raster.mask_nodata()
        .mean()
        .values,
        0.9411998,
    )

    # Use the method without gwbodies and ncfrac and waterareas from wflow
    example_wflow_model.setup_allocation_areas(
        waterareas_fn="gadm_level2",
    )
    example_wflow_model.setup_allocation_surfacewaterfrac(
        gwfrac_fn="lisflood_gwfrac",
        waterareas_fn=None,
        gwbodies_fn=None,
        ncfrac_fn=None,
        interpolate_nodata=False,
    )

    # Assert entries
    assert "demand_surface_water_ratio" in example_wflow_model.staticmaps.data
    assert np.isclose(
        example_wflow_model.staticmaps.data["demand_surface_water_ratio"]
        .raster.mask_nodata()
        .mean()
        .values,
        0.9865037,
    )


def test_setup_non_irrigation(example_wflow_model: WflowSbmModel, tmpdir: Path):
    # Read the data
    example_wflow_model.read()
    example_wflow_model.root.set(Path(tmpdir), mode="w")

    # Use the method
    example_wflow_model.setup_domestic_demand(
        domestic_fn="pcr_globwb",
        population_fn="ghs_pop_2015",
        domestic_fn_original_res=0.5,
    )
    example_wflow_model.setup_other_demand(
        demand_fn="pcr_globwb",
        variables=[
            "industry_gross",
            "industry_net",
            "livestock_gross",
            "livestock_net",
        ],
    )

    # Assert entries
    assert "demand_domestic_gross" in example_wflow_model.staticmaps.data
    assert "meta_population" in example_wflow_model.staticmaps.data

    # Assert some values
    dom_gross_vals = (
        example_wflow_model.staticmaps.data["demand_domestic_gross"]
        .isel(latitude=32, longitude=26)
        .values
    )
    assert int(np.mean(dom_gross_vals) * 100) == 127
    popu_val = (
        example_wflow_model.staticmaps.data["meta_population"]
        .isel(latitude=32, longitude=26)
        .values
    )
    assert int(popu_val) == 7450

    ind_mean = (
        example_wflow_model.staticmaps.data["demand_industry_gross"].mean().values
    )
    assert np.isclose(ind_mean, 0.065195)

    # test with other method
    example_wflow_model.setup_domestic_demand_from_population(
        population_fn="ghs_pop_2015",
        domestic_gross_per_capita=0.35,
        domestic_net_per_capita=0.25,
    )

    # Assert entries
    assert "demand_domestic_gross" in example_wflow_model.staticmaps.data
    assert "meta_population" in example_wflow_model.staticmaps.data

    # Assert some values
    dom_gross_vals = (
        example_wflow_model.staticmaps.data["demand_domestic_gross"]
        .isel(latitude=32, longitude=26)
        .values
    )
    assert int(dom_gross_vals * 100) == 109

    # Check other combination
    with pytest.raises(ValueError, match="The provided domestic demand data is "):
        example_wflow_model.setup_domestic_demand_from_population(
            population_fn="ghs_pop_2015",
            domestic_gross_per_capita=[0.35, 0.35, 0.34, 0.36],
            domestic_net_per_capita=[0.25, 0.25, 0.24, 0.26],
        )

    # Check cyclic
    example_wflow_model.setup_domestic_demand_from_population(
        population_fn="ghs_pop_2015",
        domestic_gross_per_capita=[
            0.35,
            0.35,
            0.34,
            0.34,
            0.36,
            0.36,
            0.35,
            0.35,
            0.34,
            0.34,
            0.36,
            0.36,
        ],
        domestic_net_per_capita=[
            0.25,
            0.25,
            0.24,
            0.24,
            0.26,
            0.26,
            0.25,
            0.25,
            0.24,
            0.24,
            0.26,
            0.26,
        ],
    )
    assert "time" in example_wflow_model.staticmaps.data["demand_domestic_gross"].dims
    assert "time" in example_wflow_model.staticmaps.data["demand_domestic_net"].dims


def test_setup_irrigation_nopaddy(
    example_wflow_model: WflowSbmModel, tmpdir: Path, globcover_gdf: gpd.GeoDataFrame
):
    # Read the data
    example_wflow_model.read()
    example_wflow_model.root.set(Path(tmpdir), mode="w")

    # Use the method
    example_wflow_model.setup_irrigation(
        irrigated_area_fn="irrigated_area",
        irrigation_value=[1],
        cropland_class=[11, 14, 20, 30],
        paddy_class=[],
        area_threshold=0.6,
        lai_threshold=0.2,
    )

    # Set to shorter name to improve readability of tests
    ds = example_wflow_model.staticmaps.data

    # Assert entries
    assert "demand_paddy_irrigated_mask" not in ds
    assert "demand_nonpaddy_irrigated_mask" in ds
    assert "demand_nonpaddy_irrigation_trigger" in ds

    # Assert the irrigation_trigger map has the same shape as LAI
    assert (
        ds["demand_nonpaddy_irrigation_trigger"].shape[0]
        == ds["vegetation_leaf_area_index"].shape[0]
    )

    # There is no paddy in this region
    assert ds["demand_nonpaddy_irrigated_mask"].raster.mask_nodata().sum().values == 4
    # Check if more irrigation is allowed during summer than winter
    assert (
        ds["demand_nonpaddy_irrigation_trigger"]
        .raster.mask_nodata()
        .sel(time=2)
        .sum()
        .values
        < ds["demand_nonpaddy_irrigation_trigger"]
        .raster.mask_nodata()
        .sel(time=8)
        .sum()
        .values
    )

    # Test with geodataframe
    # Use globcover gdf class 11
    irrigation_gdf = globcover_gdf.loc[globcover_gdf["landuse"].isin([14])]
    example_wflow_model.setup_irrigation_from_vector(
        irrigated_area_fn=irrigation_gdf,
        cropland_class=[11, 14, 20, 30],
        paddy_class=[],
        area_threshold=0.6,
        lai_threshold=0.2,
    )

    # Set to shorter name to improve readability of tests
    ds = example_wflow_model.staticmaps.data

    # Assert entries
    assert "demand_paddy_irrigated_mask" not in ds
    assert "demand_nonpaddy_irrigated_mask" in ds
    assert "demand_nonpaddy_irrigation_trigger" in ds

    assert ds["demand_nonpaddy_irrigated_mask"].raster.mask_nodata().sum().values == 8


def test_setup_irrigation_withpaddy(example_wflow_model: WflowSbmModel, tmpdir: Path):
    # Read the data
    example_wflow_model.read()
    example_wflow_model.root.set(
        Path(
            tmpdir,
        ),
        mode="w",
    )

    # First update landuse to add rice
    layers = [50, 100, 50, 200, 800]
    # Note: 11 this not NOT a rice class, but chosen just for testing purposes
    example_wflow_model.setup_lulcmaps_with_paddy(
        lulc_fn="glcnmo",
        paddy_class=11,
        wflow_thicknesslayers=layers,
    )

    example_wflow_model.setup_irrigation(
        irrigated_area_fn="irrigated_area",
        irrigation_value=[1],
        cropland_class=[11, 13],
        paddy_class=[11],
        area_threshold=0.6,
        lai_threshold=0.2,
    )

    # Set to shorter name to improve readability of tests
    ds = example_wflow_model.staticmaps.data

    # Assert entries
    assert "demand_paddy_irrigated_mask" in ds
    assert "demand_paddy_irrigation_trigger" in ds


def test_setup_cold_states(example_wflow_model: WflowSbmModel, tmpdir: Path):
    # Create states
    example_wflow_model.setup_cold_states()
    states = example_wflow_model.states.data.copy()

    assert "land_instantaneous_q" in example_wflow_model.states.data
    assert "layer" in example_wflow_model.states.data["soil_unsaturated_depth"].dims
    assert np.isclose(
        example_wflow_model.states.data["soil_saturated_depth"]
        .raster.mask_nodata()
        .mean()
        .values,
        648.43677,
    )
    assert np.isclose(
        example_wflow_model.states.data["subsurface_q"]
        .raster.mask_nodata()
        .mean()
        .values,
        67.45569,
    )

    # test write
    example_wflow_model.root.set(tmpdir / "wflow_cold_states", mode="w")
    example_wflow_model.states.write()

    assert (tmpdir / "wflow_cold_states" / "instate" / "instates.nc").exists()

    # test read
    example_wflow_model.root.set(tmpdir / "wflow_cold_states", mode="r+")
    example_wflow_model.states.read()

    xrt.assert_equal(
        xr.merge(states.values(), compat="override"),
        xr.merge(example_wflow_model.states.data.values(), compat="override"),
    )
